"""Calculators module for AsusRouter"""

from __future__ import annotations

from asusrouter.const import (
    CONST_PERCENTS,
    CONST_ZERO,
    DATA_TOTAL,
    DATA_USAGE,
    DATA_USED,
    DEFAULT_USAGE_DIGITS,
    ERROR_ZERO_DIVISION,
)
from asusrouter.error import AsusRouterValueError

DEFAULT_USAGE_NONE = {
    DATA_TOTAL: 0,
    DATA_USED: 0,
}


def usage(
    current_used: (int | float),
    current_total: (int | float),
    previous_used: (int | float) = CONST_ZERO,
    previous_total: (int | float) = CONST_ZERO,
) -> float:
    """Calculate usage in percents"""

    if current_used == previous_used:
        return CONST_ZERO

    if current_total == previous_total:
        raise AsusRouterValueError(
            ERROR_ZERO_DIVISION.format("current_total == previous_total")
        )

    used = current_used - previous_used
    total = current_total - previous_total

    if used < 0:
        raise AsusRouterValueError("Usage cannot be negative, used = {}".format(used))
    if total < 0:
        raise AsusRouterValueError("Usage cannot be negative, total = {}".format(total))

    if used > total:
        raise AsusRouterValueError(
            "Usage cannot be above 100%, used = {}, total = {}".format(used, total)
        )

    return round(
        CONST_PERCENTS
        * (current_used - previous_used)
        / (current_total - previous_total),
        DEFAULT_USAGE_DIGITS,
    )


def usage_in_dict(
    after: dict[str, (int | float)],
    before: dict[str, (int | float)] = DEFAULT_USAGE_NONE,
) -> dict[str, (int | float)]:
    """Calculate usage in percents in a dictionary"""

    after[DATA_USAGE] = usage(
        after[DATA_USED], after[DATA_TOTAL], before[DATA_USED], before[DATA_TOTAL]
    )

    return after


def speed(
    after: (int | float),
    before: (int | float),
    time_delta: (int | float) | None = None,
) -> float:
    """Calculate speed"""

    if time_delta is None:
        return CONST_ZERO
    elif time_delta == CONST_ZERO:
        raise AsusRouterValueError(ERROR_ZERO_DIVISION.format("time_delta"))

    diff = after - before if after > before else CONST_ZERO

    return diff / time_delta


def rgb(raw: dict[int, dict[str, int]]) -> dict[int, dict[str, int]]:
    """Calculate RGB values from input"""

    output = dict()

    for led in raw:
        output[led] = dict()
        for channel in raw[led]:
            value = raw[led][channel]
            if value < 0:
                output[led][channel] = 0
            elif value > 128:
                output[led][channel] = 128
            else:
                output[led][channel] = value

    return output



"""Connection module"""

from __future__ import annotations

import asyncio
import base64
import json
import logging
import ssl
import urllib.parse
from pathlib import Path
from typing import Any

import aiohttp

from asusrouter import (
    AsusRouter404,
    AsusRouterAuthorizationError,
    AsusRouterConnectionError,
    AsusRouterConnectionTimeoutError,
    AsusRouterError,
    AsusRouterLoginBlockError,
    AsusRouterLoginError,
    AsusRouterResponseError,
    AsusRouterServerDisconnectedError,
    AsusRouterSSLError,
)
from asusrouter.const import (
    AR_API,
    AR_ERROR,
    AR_PATH,
    AR_USER_AGENT,
    DEFAULT_PORT,
    DEFAULT_SLEEP_CONNECTING,
    DEFAULT_SLEEP_RECONNECT,
    MSG_ERROR,
    MSG_INFO,
    MSG_SUCCESS,
    MSG_WARNING,
)
from asusrouter.util import converters, parsers

_LOGGER = logging.getLogger(__name__)


class Connection:
    """Create connection"""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        port: int | None = None,
        use_ssl: bool = False,
        cert_check: bool = True,
        cert_path: str = "",
        session: aiohttp.ClientSession | None = None,
    ):
        """Properties for connection"""

        self._host: str | None = host
        self._port: int | None = port
        self._username: str | None = username
        self._password: str | None = password
        self._token: str | None = None
        self._headers: dict[str, str] | None = None
        self._session: aiohttp.ClientSession | None = session

        self._device: dict[str, str] = dict()
        self._error: bool = False
        self._connecting: bool = False
        self._mute_flag: bool = False
        self._drop: bool = False

        self._http = "https" if use_ssl else "http"

        if self._port is None or self._port == 0:
            self._port = DEFAULT_PORT[self._http]

        if cert_check:
            if cert_path != "":
                path = Path(cert_path)
                if path.is_file():
                    self._ssl = ssl.create_default_context(cafile=cert_path)
                    _LOGGER.debug(MSG_SUCCESS["cert_found"])
                else:
                    _LOGGER.error(MSG_ERROR["cert_missing"])
                    self._ssl = True
            else:
                _LOGGER.debug(MSG_INFO["no_cert"])
                self._ssl = True
        else:
            _LOGGER.debug(MSG_INFO["no_cert_check"])
            self._ssl = False

        self._connected: bool = None

    async def async_run_command(
        self, command: str, endpoint: str = AR_PATH["get"], retry: bool = False
    ) -> dict[str, Any]:
        """Run command. Use the existing connection token, otherwise create new one"""

        if self._drop:
            return {}

        if self._token is None and not retry:
            _LOGGER.debug(f"No token - connecting and repeating")
            await self.async_connect()
            return await self.async_run_command(command, endpoint, retry=True)
        else:
            if self._token is not None:
                try:
                    result = await self.async_request(command, endpoint, self._headers)
                    return result
                except AsusRouterAuthorizationError as ex:
                    if not retry:
                        await self.async_connect()
                        return await self.async_run_command(
                            command, endpoint, retry=True
                        )
                    else:
                        raise ex
                except AsusRouterError as ex:
                    raise ex
                except Exception as ex:
                    if not retry:
                        await self.async_connect()
                        return await self.async_run_command(
                            command, endpoint, retry=True
                        )
                    else:
                        raise ex
            else:
                _LOGGER.error(MSG_ERROR["command"].format(endpoint))
                return {}

    async def async_load(self, endpoint: str, retry: bool = False) -> str:
        """Load the page"""

        if self._drop:
            return {}

        if self._token is None and not retry:
            await self.async_connect()
            return await self.async_load(endpoint, retry=True)

        string_body = str()

        try:
            async with self._session.post(
                url="{}://{}:{}/{}".format(
                    self._http, self._host, self._port, endpoint
                ),
                headers=self._headers,
                ssl=self._ssl,
            ) as r:
                string_body = await r.text()
                if "404 Not Found" in string_body:
                    raise AsusRouter404()
        except (aiohttp.ClientError, asyncio.TimeoutError) as ex:
            if not retry:
                return self.async_load(endpoint, retry=True)
            raise AsusRouterConnectionError(ex)
        except Exception as ex:
            raise ex

        return string_body

    async def async_request(
        self, payload: str, endpoint: str, headers: dict[str, str], retry: bool = False
    ) -> dict[str, Any]:
        """Send a request"""

        if self._drop:
            return {}

        # Wait a bit before just retrying
        if retry:
            await asyncio.sleep(DEFAULT_SLEEP_RECONNECT)

        # Connection process
        if endpoint == AR_PATH["login"]:
            self._connecting = True
        # Sleep if we got here during the connection process
        else:
            while self._connecting:
                await asyncio.sleep(DEFAULT_SLEEP_CONNECTING)

        json_body = {}

        try:
            async with self._session.post(
                url="{}://{}:{}/{}".format(
                    self._http, self._host, self._port, endpoint
                ),
                data=urllib.parse.quote(payload),
                headers=headers,
                ssl=self._ssl,
            ) as r:
                string_body = await r.text()
                if string_body == str():
                    return dict()
                if "404 Not Found" in string_body:
                    raise AsusRouter404()
                json_body = await r.json()

                # Handle reported errors
                if "error_status" in json_body:
                    # If got here, we are not connected properly
                    self._connected = False

                    ## ERROR CODES
                    error_code = int(json_body["error_status"])
                    # Not authorised
                    if error_code == AR_ERROR["authorisation"]:
                        raise AsusRouterAuthorizationError(MSG_ERROR["authorisation"])
                    # Wrong crerdentials
                    elif error_code == AR_ERROR["credentials"]:
                        raise AsusRouterLoginError(MSG_ERROR["credentials"])
                    # Too many attempts
                    elif error_code == AR_ERROR["try_again"]:
                        raise AsusRouterLoginBlockError(
                            MSG_ERROR["try_again"],
                            timeout=converters.int_from_str(
                                json_body["remaining_lock_time"]
                            ),
                        )
                    # Loged out
                    elif error_code == AR_ERROR["logout"]:
                        return {"success": True}
                    # Unknown error code
                    else:
                        raise AsusRouterResponseError(
                            MSG_ERROR["unknown"].format(error_code)
                        )

            # If loged in, save the device API data
            if endpoint == AR_PATH["login"]:
                r_headers = r.headers
                for item in AR_API:
                    if item in r_headers:
                        self._device[item] = r_headers[item]

            # Reset mute_flag on success
            self._mute_flag = False

            return json_body

        # Handle non-JSON replies
        except json.JSONDecodeError:
            if ".xml" in endpoint:
                json_body = parsers.xml(text=string_body, page=endpoint)
            else:
                json_body = parsers.pseudo_json(text=string_body, page=endpoint)
            return json_body

        # Raise only if mute_flag not set
        except (
            aiohttp.ClientConnectorSSLError,
            aiohttp.ClientConnectorCertificateError,
        ) as ex:
            if not self._mute_flag:
                raise AsusRouterSSLError()
        except (aiohttp.ServerTimeoutError, asyncio.TimeoutError) as ex:
            if not self._mute_flag:
                raise AsusRouterConnectionTimeoutError()
        # This happens regularly if more connections are established. Repeat before actually raising the exception
        except aiohttp.ServerDisconnectedError as ex:
            if retry:
                raise AsusRouterServerDisconnectedError(ex) from ex
            else:
                _LOGGER.debug(MSG_WARNING["disconnected"].format(endpoint, payload))

        # Connection refused or reset by peer -> will repeat
        except (aiohttp.ClientOSError) as ex:
            if endpoint == AR_PATH["login"] and not self._mute_flag:
                raise AsusRouterConnectionError(str(ex)) from ex
            # Mute warning for retries and while connecting
            if not retry and not self._connecting:
                _LOGGER.debug(
                    "{}. Endpoint: {}. Payload: {}.\nException summary:{}".format(
                        MSG_WARNING["refused"], endpoint, payload, str(ex)
                    )
                )

        # If it got here, something is wrong. Reconnect and retry
        self._mute_flag = True

        if not retry:
            self._error = True
            _LOGGER.debug(MSG_INFO["reconnect"])
            await self.async_cleanup()
            await self.async_connect(retry=True)
        return await self.async_request(
            payload=payload, endpoint=endpoint, headers=headers, retry=True
        )

    async def async_connect(self, retry: bool = False) -> bool:
        """Start new connection to and get new auth token"""

        _success = False
        self._drop = False

        if self._session is None:
            self._session = aiohttp.ClientSession()

        auth = "{}:{}".format(self._username, self._password).encode("ascii")
        logintoken = base64.b64encode(auth).decode("ascii")
        payload = "login_authorization={}".format(logintoken)
        headers = {"user-agent": AR_USER_AGENT}

        response = await self.async_request(
            payload=payload, endpoint=AR_PATH["login"], headers=headers, retry=retry
        )
        if "asus_token" in response:
            self._token = response["asus_token"]
            self._headers = {
                "user-agent": AR_USER_AGENT,
                "cookie": "asus_token={}".format(self._token),
            }
            _LOGGER.debug(
                "{} on port {}: {}".format(
                    MSG_SUCCESS["login"], self._port, self._device
                )
            )

            self._connected = True
            _success = True
        else:
            _LOGGER.error(MSG_ERROR["token"])

        self._connecting = False

        return _success

    async def async_disconnect(self) -> bool:
        """Close the connection"""

        # Not connected
        if not self._connected:
            _LOGGER.debug(MSG_WARNING["not_connected"])
        # Connected
        else:
            result = await self.async_request("", AR_PATH["logout"], self._headers)
            if not "success" in result:
                return False

        # Clean up
        await self.async_cleanup()

        return True

    async def async_drop_connection(self) -> None:
        """Drop connection"""

        self._drop = True
        await self.async_cleanup()

    async def async_cleanup(self) -> None:
        """Cleanup after logout"""

        self._connected = False
        self._token = None
        self._headers = None

    async def async_reset_error(self) -> None:
        """Reset error flag"""

        self._error = False
        return

    @property
    def connected(self) -> bool:
        """Connection status"""

        return self._connected

    @property
    def connecting(self) -> bool:
        """Connection progress"""

        return self._connecting

    @property
    def device(self) -> dict[str, str]:
        """Device model and API support levels"""

        return self._device

    @property
    def error(self) -> bool:
        """Report errors"""

        return self._error



"""Compilers module for AsusRouter"""

from __future__ import annotations
from dataclasses import asdict
from typing import Any

from asusrouter.const import (
    AR_DEFAULT_OVPN_CLIENTS,
    AR_HOOK_TEMPLATE,
    AR_KEY_OVPN,
    AR_KEY_PARENTAL_CONTROL_MAC,
    AR_KEY_PARENTAL_CONTROL_NAME,
    AR_KEY_PARENTAL_CONTROL_STATE,
    AR_KEY_PARENTAL_CONTROL_TIMEMAP,
    AR_KEY_VPN_CLIENT,
    AR_MAP_PARENTAL_CONTROL_STATE,
    AR_MAP_RGB,
    AR_VPN_STATUS,
    ENDPOINT,
    ENDPOINT_ARGS,
    ERROR_VALUE_TYPE,
    KEY_NVRAM_GET,
    KEY_VPN,
    PARAM_ERRNO,
    PARAM_STATE,
    PARAM_STATUS,
    PARAM_UNKNOWN,
)
from asusrouter import FilterDevice
from asusrouter.dataclass import AsusDevice, ConnectedDevice
from asusrouter.error import AsusRouterValueError
from .converters import int_from_str


def hook(commands: dict[str, str] | None = None) -> str:
    """Hook compiler"""

    data = str()
    if commands is not None:
        for item in commands:
            data += AR_HOOK_TEMPLATE.format(item, commands[item])

    return data


def nvram(values: list[str] | str | None = None) -> str:
    """NVRAM request compiler"""

    if values is None:
        return str()

    if type(values) == str:
        return AR_HOOK_TEMPLATE.format(KEY_NVRAM_GET, values)

    request = str()
    for value in values:
        request += AR_HOOK_TEMPLATE.format(KEY_NVRAM_GET, value)

    return request


def rgb(raw: dict[int, dict[str, int]]) -> str:
    """RGB value compiler for LEDG scheme"""

    value = str()

    raw = dict(sorted(raw.items()))

    for led in raw:
        for channel in AR_MAP_RGB:
            if AR_MAP_RGB[channel] in raw[led]:
                value += f"{raw[led][AR_MAP_RGB[channel]]},"
            else:
                value += "0,"

    value = value[:-1]

    return value


def vpn_from_devicemap(
    vpn: dict[str, Any], devicemap: dict[str, Any]
) -> dict[str, Any]:
    """Compile devicemap into VPN"""

    if type(devicemap) != dict:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(devicemap, type(devicemap)))

    if type(vpn) != dict:
        vpn = dict()

    if KEY_VPN in devicemap:
        for num in range(1, AR_DEFAULT_OVPN_CLIENTS + 1):
            key = f"{AR_KEY_VPN_CLIENT}{num}"
            if not key in vpn:
                vpn[key] = dict()
            if (
                AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_STATE)
                in devicemap[KEY_VPN]
            ):
                vpn[key][PARAM_STATUS] = int_from_str(
                    devicemap[KEY_VPN][
                        AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_STATE)
                    ]
                )
                if vpn[key][PARAM_STATUS] == 1 or vpn[key][PARAM_STATUS] == 2:
                    vpn[key][PARAM_STATE] = True
                else:
                    vpn[key][PARAM_STATE] = False
                if vpn[key][PARAM_STATUS] in AR_VPN_STATUS:
                    vpn[key][PARAM_STATUS] = AR_VPN_STATUS[vpn[key][PARAM_STATUS]]
                else:
                    vpn[key][
                        PARAM_STATUS
                    ] = f"{PARAM_UNKNOWN} ({vpn[key][PARAM_STATUS]})"
            if (
                AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_ERRNO)
                in devicemap[KEY_VPN]
            ):
                vpn[key][PARAM_ERRNO] = int_from_str(
                    devicemap[KEY_VPN][
                        AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_ERRNO)
                    ]
                )

    return vpn


def parental_control(data: dict[str, FilterDevice]) -> dict[str, str]:
    """Compile parental control rules"""

    if type(data) != dict:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(data, type(data)))

    state_lib = dict()
    for index, state in AR_MAP_PARENTAL_CONTROL_STATE.items():
        state_lib[state] = index

    macs = str()
    names = str()
    states = str()
    timemaps = str()

    for rule in data:
        macs += data[rule].mac + ">"
        names += data[rule].name + ">"
        states += state_lib[data[rule].state] + ">"
        timemaps += data[rule].timemap.replace("&#60", "<") + ">"

    macs = macs[:-1]
    names = names[:-1]
    states = states[:-1]
    timemaps = timemaps[:-1]

    return {
        AR_KEY_PARENTAL_CONTROL_MAC: macs,
        AR_KEY_PARENTAL_CONTROL_NAME: names,
        AR_KEY_PARENTAL_CONTROL_STATE: states,
        AR_KEY_PARENTAL_CONTROL_TIMEMAP: timemaps,
    }


def connected_device(
    device: ConnectedDevice, state: dict[str, Any] | None = None
) -> ConnectedDevice:
    """Compile connected device from different sources"""

    if state is None:
        device.online = False
        return device

    values: dict[str, Any] = asdict(device)
    values.update(state)

    return ConnectedDevice(**values)


def endpoint(endpoint: str, device: AsusDevice | dict[str, Any] | None = None) -> str:
    """Compile endpoint string with required arguments"""

    if endpoint not in ENDPOINT:
        raise AsusRouterValueError(f"Unknown endpoint `{endpoint}`")

    address = ENDPOINT.get(endpoint, str())

    if type(device) == AsusDevice:
        device = asdict(device)
    if type(device) != dict:
        return address

    address += "?"

    if endpoint in ENDPOINT_ARGS:
        for key in ENDPOINT_ARGS[endpoint]:
            address += f"{ENDPOINT_ARGS[endpoint][key]}={device.get(key, str())}&"

    return address



"""AsusRouter module"""

from __future__ import annotations

import aiohttp
import asyncio
from datetime import datetime
import logging
from typing import Any

from asusrouter import (
    AsusDevice,
    AsusRouter404,
    AsusRouterError,
    AsusRouterIdentityError,
    AsusRouterServerDisconnectedError,
    AsusRouterServiceError,
    AsusRouterValueError,
    Connection,
    FilterDevice,
    Monitor,
)
from asusrouter.const import (
    AIMESH,
    AR_DEVICE_IDENTITY,
    AR_FIRMWARE_CHECK_COMMAND,
    AR_HOOK_DEVICES,
    AR_KEY_AURARGB,
    AR_KEY_CPU,
    AR_KEY_DEVICES,
    AR_KEY_LED,
    AR_KEY_LEDG_COUNT,
    AR_KEY_LEDG_RGB,
    AR_KEY_LEDG_SCHEME,
    AR_KEY_LEDG_SCHEME_OLD,
    AR_KEY_NETWORK,
    AR_KEY_PARENTAL_CONTROL,
    AR_KEY_RAM,
    AR_KEY_SERVICE_COMMAND,
    AR_KEY_SERVICE_MODIFY,
    AR_KEY_SERVICE_REPLY,
    AR_KEY_WAN,
    AR_LEDG_MODE,
    AR_MAP_PARENTAL_CONTROL,
    AR_PATH,
    AR_SERVICE_COMMAND,
    AR_SERVICE_CONTROL,
    AR_SERVICE_DROP_CONNECTION,
    CLIENTS,
    CONNECTION_TYPE,
    CONVERTERS,
    DATA_BY_CORE,
    DATA_TOTAL,
    DATA_USAGE,
    DATA_USED,
    DEFAULT_ACTION_MODE,
    DEFAULT_CACHE_TIME,
    DEFAULT_SLEEP_TIME,
    ENDPOINT,
    ENDPOINTS,
    ERROR_IDENTITY,
    ERROR_SERVICE,
    ERROR_SERVICE_UNKNOWN,
    ERROR_VALUE,
    ETHERNET_PORTS,
    GUEST,
    INFO,
    INTERFACE_TYPE,
    IP,
    KEY_ACTION_MODE,
    KEY_CPU,
    KEY_HOOK,
    KEY_NETWORK,
    KEY_RAM,
    KEY_SYSINFO,
    KEY_TEMPERATURE,
    KEY_VPN,
    KEY_WAN,
    LAN,
    LINK_RATE,
    MAC,
    MONITOR_MAIN,
    MSG_ERROR,
    MSG_INFO,
    MSG_SUCCESS,
    NODE,
    NVRAM_LIST,
    NVRAM_TEMPLATE,
    ONBOARDING,
    PARAM_COLOR,
    PARAM_COUNT,
    PARAM_MODE,
    PORT,
    PORT_STATUS,
    PORT_TYPE,
    PORT_TYPES,
    PORTS,
    RSSI,
    SPEED_TYPES,
    STATE,
    SYSINFO,
    TRACK_SERVICES_LED,
    USB,
    WAN,
)
from asusrouter.dataclass import AiMeshDevice
from asusrouter.util import calculators, compilers, converters, parsers

_LOGGER = logging.getLogger(__name__)


class AsusRouter:
    """The interface class"""

    def __init__(
        self,
        host: str,
        username: str | None = None,
        password: str | None = None,
        port: int | None = None,
        use_ssl: bool = False,
        cert_check: bool = True,
        cert_path: str = "",
        cache_time: int = DEFAULT_CACHE_TIME,
        enable_monitor: bool = True,
        enable_control: bool = False,
        session: aiohttp.ClientSession | None = None,
    ):
        """Init"""

        self._host: str = host

        self._cache_time: int = cache_time
        self._enable_monitor: bool = enable_monitor
        self._enable_control: bool = enable_control

        self._device_cpu_cores: list[int] | None = None
        self._device_ports: dict[str, str] | None = None
        self._device_boottime: datetime | None = None

        self.monitor: dict[str, Monitor] = {
            ETHERNET_PORTS: Monitor(),
            ONBOARDING: Monitor(),
            PORT_STATUS: Monitor(),
            SYSINFO: Monitor(),
        }

        self._monitor_main: Monitor = Monitor()
        self._monitor_nvram: Monitor = Monitor()
        self._monitor_misc: Monitor = Monitor()
        self._monitor_devices: Monitor = Monitor()

        self._identity: AsusDevice | None = None
        self._ledg_color: dict[int, dict[str, int]] | None = None
        self._ledg_count: int = 0
        self._ledg_mode: int | None = None
        self._status_led: bool = False

        """Connect"""
        self.connection = Connection(
            host=self._host,
            username=username,
            password=password,
            port=port,
            use_ssl=use_ssl,
            cert_check=cert_check,
            cert_path=cert_path,
            session=session,
        )

    ### MAIN CONTROL -->

    async def async_check_endpoint(self, endpoint: str) -> bool:
        """Check if endpoint exists"""

        try:
            result = await self.connection.async_load(endpoint)
            return True
        except AsusRouter404:
            return False

    async def async_connect(self) -> bool:
        """Connect to the device"""

        try:
            await self.connection.async_connect()
        except Exception as ex:
            raise ex

        await self.async_identify()

        return True

    async def async_disconnect(self) -> bool:
        """Disconnect from the device"""

        try:
            await self.connection.async_disconnect()
        except Exception as ex:
            raise ex

        return True

    async def async_drop_connection(self) -> bool:
        """Drop connection when device will not reply because of our actions"""

        try:
            await self.connection.async_drop_connection()
        except Exception as ex:
            raise ex

        return True

    async def async_get_identity(self, force: bool = False) -> AsusDevice:
        """Return device identity"""

        if not self._identity or force:
            await self.async_identify()

        return self._identity

    async def async_identify(self) -> None:
        """Identify the device"""

        _LOGGER.debug(MSG_INFO["identifying"])

        # Compile
        query = list()
        for item in AR_DEVICE_IDENTITY:
            query.append(item)

        # Collect data
        message = compilers.nvram(query)
        try:
            raw = await self.async_hook(message, force=True)
        except Exception as ex:
            raise AsusRouterIdentityError(ERROR_IDENTITY.format(self._host, str(ex)))

        # Parse
        identity = dict()
        for item in AR_DEVICE_IDENTITY:
            key = item.get()
            try:
                data = item.method(raw[item.value]) if item.method else raw[item.value]
                if key in identity:
                    if type(identity[key]) is list:
                        identity[key].extend(data)
                    else:
                        identity[key] = data
                else:
                    identity[key] = data
            except Exception as ex:
                raise AsusRouterIdentityError(
                    ERROR_IDENTITY.format(self._host, str(ex))
                )

        # Mac (for some Merlin devices missing label_mac)
        if identity["mac"] is None or identity["mac"] == str():
            if identity["lan_mac"] is not None:
                identity["mac"] = identity["lan_mac"]
            elif identity["wan_mac"] is not None:
                identity["mac"] = identity["wan_mac"]

        identity.pop("lan_mac")
        identity.pop("wan_mac")

        # Firmware
        identity["firmware"] = parsers.firmware_string(
            f"{identity['fw_major']}.{identity['fw_minor']}.{identity['fw_build']}"
        )
        identity.pop("fw_major")
        identity.pop("fw_minor")
        identity.pop("fw_build")

        # Check by page
        for endpoint in ENDPOINT:
            self.monitor[endpoint].enabled = identity[
                endpoint
            ] = await self.async_check_endpoint(ENDPOINT[endpoint])
        identity["update_networkmapd"] = await self.async_check_endpoint(
            endpoint=AR_PATH["networkmap"]
        )

        # Check RGBG / AURA
        try:
            data = await self.connection.async_load(AR_PATH["rgb"])
            if AR_KEY_AURARGB in data:
                identity["aura"] = True
            elif AR_KEY_LEDG_COUNT in data:
                identity["ledg"] = True
                self._ledg_count = parsers.ledg_count(data)
        except AsusRouter404 as ex:
            """Do nothing"""

        # Check VPN
        try:
            data = await self.connection.async_load(AR_PATH["vpn"])
            identity["vpn_status"] = True
        except AsusRouter404 as ex:
            """Do nothing"""

        # Save static values
        self._status_led = raw[AR_KEY_LED]

        self._identity = AsusDevice(**identity)

        _LOGGER.debug(MSG_SUCCESS["identity"])

    ### <-- MAIN CONTROL

    ### DATA PROCESSING -->

    async def async_command(
        self,
        commands: dict[str, str] | None = None,
        action_mode: str = DEFAULT_ACTION_MODE,
        apply_to: str = AR_PATH["command"],
    ) -> dict[str, Any]:
        """Command device to run a service or set parameter"""

        result = {}

        if not self._enable_control:
            _LOGGER.error(MSG_ERROR["disabled_control"])
            return result

        request: dict = {
            KEY_ACTION_MODE: action_mode,
        }
        if commands is not None:
            for command in commands:
                request[command] = commands[command]

        try:
            result = await self.connection.async_run_command(
                command=str(request), endpoint=apply_to
            )
        except Exception as ex:
            raise ex

        return result

    async def async_hook(
        self, hook: str | None = None, force: bool = False
    ) -> dict[str, Any]:
        """Hook data from device"""

        result = {}

        if not self._enable_monitor and not force:
            _LOGGER.error(MSG_ERROR["disabled_monitor"])
            return result

        if hook is None:
            _LOGGER.debug(MSG_INFO["empty_request"])
            return result

        try:
            result = await self.connection.async_run_command(
                command="{}={}".format(KEY_HOOK, hook)
            )
        except Exception as ex:
            raise ex

        # Check for errors during hook
        if self.connection.error:
            _LOGGER.debug(MSG_INFO["error_flag"])
            await self.async_handle_error()

        return result

    async def async_load(self, page: str | None = None) -> dict[str, Any]:
        """Return the data from the page"""

        result = {}

        if not self._enable_monitor:
            _LOGGER.error(MSG_ERROR["disabled_monitor"])
            return result

        if page is None:
            _LOGGER.debug(MSG_INFO["empty_request"])
            return result

        try:
            result = await self.connection.async_run_command(command="", endpoint=page)
        except AsusRouter404 as ex:
            raise ex
        except Exception as ex:
            raise ex

        return result

    ### <-- DATA PROCESSING

    ### MONITORS -->

    async def async_monitor_main(self) -> None:
        """Get all the main monitor values. Non-cacheable"""

        while self._monitor_main.active:
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            return

        try:
            self._monitor_main.start()

            monitor_main = Monitor()

            hook = compilers.hook(MONITOR_MAIN)
            data = await self.async_hook(hook)

            monitor_main.reset()

            ### CPU ###
            ## Not yet known what exactly is type of data. But this is the correct way to calculate usage from it

            # Keep safe
            if AR_KEY_CPU in data:

                raw_cpu = data[AR_KEY_CPU]

                # Check which cores CPU has or find it out and save for later
                if self._device_cpu_cores is None:
                    self._device_cpu_cores = parsers.cpu_cores(raw=raw_cpu)

                # Traffic data from the device
                monitor_main[KEY_CPU] = parsers.cpu_usage(
                    raw=raw_cpu, cores=self._device_cpu_cores
                )

                # Calculate actual usage in percents and save it. Only if there was old data for CPU
                if self._monitor_main.ready and KEY_CPU in self._monitor_main:
                    for item in monitor_main[KEY_CPU]:
                        if item in self._monitor_main[KEY_CPU]:
                            monitor_main[KEY_CPU][item] = calculators.usage_in_dict(
                                after=monitor_main[KEY_CPU][item],
                                before=self._monitor_main[KEY_CPU][item],
                            )

            # Keep last data
            elif self._monitor_main.ready and KEY_CPU in self._monitor_main:
                monitor_main[KEY_CPU] = self._monitor_main[KEY_CPU]

            ### RAM ###
            ## Data is in KiB. To get MB as they are shown in the device Web-GUI, should be devided by 1024 (yes, those will be MiB)

            if AR_KEY_RAM in data:
                # Populate RAM with known values
                monitor_main[KEY_RAM] = parsers.ram_usage(raw=data[AR_KEY_RAM])

                # Calculate usage in percents
                if (
                    DATA_USED in monitor_main[KEY_RAM]
                    and DATA_TOTAL in monitor_main[KEY_RAM]
                ):
                    monitor_main[KEY_RAM] = calculators.usage_in_dict(
                        after=monitor_main[KEY_RAM]
                    )
            # Keep last data
            elif self._monitor_main.ready and KEY_RAM in self._monitor_main:
                monitor_main[KEY_RAM] = self._monitor_main[KEY_RAM]

            ### NETWORK ###
            ## Data in Bytes for traffic and bits/s for speeds

            if AR_KEY_NETWORK in data:
                # Calculate RX and TX from the HEX values. If there is no current value, but there was one before, get it from storage. Traffic resets only on device reboot or when above the limit. Device disconnect / reconnect does NOT reset it
                monitor_main[KEY_NETWORK] = parsers.network_usage(
                    raw=data[AR_KEY_NETWORK]
                )

                if self._monitor_main.ready and KEY_NETWORK in self._monitor_main:
                    # Calculate speeds
                    time_delta = (
                        monitor_main.time - self._monitor_main.time
                    ).total_seconds()
                    monitor_main[KEY_NETWORK] = parsers.network_speed(
                        after=monitor_main[KEY_NETWORK],
                        before=self._monitor_main[KEY_NETWORK],
                        time_delta=time_delta,
                    )

            # Keep last data
            elif self._monitor_main.ready and KEY_NETWORK in self._monitor_main:
                monitor_main[KEY_NETWORK] = self._monitor_main[KEY_NETWORK]

            ### WAN STATE ###

            if AR_KEY_WAN in data:
                # Populate WAN with known values
                monitor_main[KEY_WAN] = parsers.wan_state(raw=data[AR_KEY_WAN])

            # Keep last data
            elif self._monitor_main.ready and KEY_WAN in self._monitor_main:
                monitor_main[KEY_WAN] = self._monitor_main[KEY_WAN]

            ### SYSINFO ###

            if self._identity:
                if self._identity.sysinfo:
                    monitor_main[KEY_SYSINFO] = await self.async_load(
                        page=AR_PATH["sysinfo"]
                    )

            monitor_main.finish()
            self._monitor_main = monitor_main
        except AsusRouterError as ex:
            self._monitor_main.drop()
            raise ex

        return

    async def async_monitor_nvram(self, groups: list[str] | str | None = None) -> None:
        """Get the NVRAM values for the specified group list"""

        while self._monitor_nvram.active:
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            return

        try:
            self._monitor_nvram.start()

            monitor_nvram = Monitor()

            # If none groups were sent, will return all the known NVRAM values
            if groups is None:
                groups = [*NVRAM_LIST]

            # If only one group is sent
            if type(groups) is not list:
                groups = [groups.upper()]

            requests = []
            for group in groups:
                group = group.upper()
                if group in NVRAM_LIST:
                    for value in NVRAM_LIST[group]:
                        requests.append(value)
                else:
                    _LOGGER.warning(
                        "There is no {} in known NVRAM groups".format(group)
                    )

            message = compilers.nvram(requests)
            result = await self.async_hook(message)

            monitor_nvram.reset()

            for item in result:
                monitor_nvram[item] = result[item]

            monitor_nvram.finish()
            self._monitor_nvram = monitor_nvram
        except AsusRouterError as ex:
            self._monitor_nvram.drop()
            raise ex

        return

    async def async_monitor_misc(self) -> None:
        """Get all other useful values"""

        while self._monitor_misc.active:
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            return

        try:
            self._monitor_misc.start()

            monitor_misc = Monitor()

            ### PORTS

            # Receive ports number and status (disconnected, 100 Mb/s, 1 Gb/s)
            data = await self.async_load(page=AR_PATH["ports"])

            monitor_misc.reset()

            monitor_misc["PORTS"] = dict()
            if "portSpeed" in data:
                for type in PORT_TYPE:
                    monitor_misc["PORTS"][type] = dict()
                for value in data["portSpeed"]:
                    for type in PORT_TYPE:
                        if type in value:
                            number = value.replace(type, "")
                            monitor_misc["PORTS"][type][number] = parsers.port_speed(
                                data["portSpeed"][value]
                            )
                            break

            ### TEMPERATURES ###
            monitor_misc[KEY_TEMPERATURE] = await self.async_load(
                page=AR_PATH["temperature"]
            )

            # Load devicemap
            data = await self.async_load(page=AR_PATH["devicemap"])
            monitor_misc["DEVICEMAP"] = data

            ### VPN ###
            if self._identity.vpn_status:
                monitor_misc[KEY_VPN] = await self.async_load(page=AR_PATH["vpn"])
                if monitor_misc["DEVICEMAP"] and monitor_misc[KEY_VPN]:
                    monitor_misc[KEY_VPN] = compilers.vpn_from_devicemap(
                        monitor_misc[KEY_VPN], monitor_misc["DEVICEMAP"]
                    )
            else:
                monitor_misc[KEY_VPN] = compilers.vpn_from_devicemap(
                    None, monitor_misc["DEVICEMAP"]
                )

            # Calculate boot time. Since precision is 1 second, could be that old and new are 1 sec different. In this case, we should not change the boot time, but keep the previous value to avoid regular changes
            if "SYS" in monitor_misc["DEVICEMAP"]:
                if "uptimeStr" in monitor_misc["DEVICEMAP"]["SYS"]:
                    time = parsers.uptime(monitor_misc["DEVICEMAP"]["SYS"]["uptimeStr"])
                    timestamp = int(time.timestamp())
                    if not "BOOTTIME" in monitor_misc:
                        monitor_misc["BOOTTIME"] = dict()

                    if "BOOTTIME" in self._monitor_misc:
                        _timestamp = self._monitor_misc["BOOTTIME"]["timestamp"]
                        _time = datetime.fromtimestamp(_timestamp)
                        if time == _time or abs(timestamp - _timestamp) < 2:
                            monitor_misc["BOOTTIME"] = self._monitor_misc["BOOTTIME"]
                        # This happens on reboots if data is checked before the device got correct time from NTP
                        elif timestamp - _timestamp < 0:
                            # Leave the same boot time, since we don't know the new correct time
                            monitor_misc["BOOTTIME"] = self._monitor_misc["BOOTTIME"]
                        # Boot time changed -> there was reboot
                        else:
                            monitor_misc["BOOTTIME"]["timestamp"] = timestamp
                            monitor_misc["BOOTTIME"]["ISO"] = time.isoformat()
                            await self.async_handle_reboot()
                    else:
                        monitor_misc["BOOTTIME"]["timestamp"] = timestamp
                        monitor_misc["BOOTTIME"]["ISO"] = time.isoformat()

                    if self._device_boottime != monitor_misc["BOOTTIME"]["ISO"]:
                        self._device_boottime = monitor_misc["BOOTTIME"]["ISO"]

            monitor_misc.finish()
            self._monitor_misc = monitor_misc
        except AsusRouterError as ex:
            self._monitor_misc.drop()
            raise ex

        return

    async def async_monitor_ethernet_ports(self) -> None:
        """Monitor ethernet ports state"""

        # Check whether to run
        if not self.async_monitor_should_run(ETHERNET_PORTS):
            return

        try:
            # Start
            self.monitor[ETHERNET_PORTS].start()
            monitor = Monitor()
            # Hook data
            raw = await self.async_load(
                compilers.endpoint(ETHERNET_PORTS, self._identity)
            )
            # Reset time
            monitor.reset()

            # Process data

            # Ports info
            ports = {
                LAN: dict(),
                WAN: dict(),
            }
            data = raw["portSpeed"]
            for port in data:
                port_type = port[0:3].lower()
                port_id = converters.int_from_str(port[3:])
                value = SPEED_TYPES.get(data[port])
                ports[port_type][port_id] = {
                    STATE: converters.bool_from_any(value),
                    LINK_RATE: value,
                }

            monitor[PORTS] = ports

            # Finish and save data
            monitor.finish()
            self.monitor[ETHERNET_PORTS] = monitor

        except AsusRouterError as ex:
            self.monitor[ETHERNET_PORTS].drop()
            raise ex

        return

    async def async_monitor_onboarding(self) -> None:
        """Get the onboarding state"""

        # Check whether to run
        if not self.async_monitor_should_run(ONBOARDING):
            return

        try:
            # Start
            self.monitor[ONBOARDING].start()
            monitor = Monitor()
            # Hook data
            raw = await self.async_load(AR_PATH["onboarding"])
            # Reset time
            monitor.reset()

            # Process data

            # AiMesh nodes state
            aimesh = dict()
            data = raw["get_cfg_clientlist"][0]
            for device in data:
                node = parsers.aimesh_node(device)
                aimesh[node.mac] = node
            monitor[AIMESH] = aimesh

            # Client list
            clients = dict()
            data = raw["get_allclientlist"][0]
            for node in data:
                for connection in data[node]:
                    for mac in data[node][connection]:
                        convert = converters.onboarding_connection(connection)
                        description = {
                            CONNECTION_TYPE: convert[CONNECTION_TYPE],
                            GUEST: convert[GUEST],
                            IP: converters.none_or_str(
                                data[node][connection][mac].get(IP, None)
                            ),
                            MAC: mac,
                            NODE: node,
                            RSSI: data[node][connection][mac].get(RSSI, None),
                        }
                        clients[mac] = description
            monitor[CLIENTS] = clients

            # Finish and save data
            monitor.finish()
            self.monitor[ONBOARDING] = monitor

        except AsusRouterError as ex:
            self.monitor[ONBOARDING].drop()
            raise ex

        return

    async def async_monitor_port_status(self) -> None:
        """Monitor port status for the node"""

        # Check whether to run
        if not self.async_monitor_should_run(PORT_STATUS):
            return

        try:
            # Start
            self.monitor[PORT_STATUS].start()
            monitor = Monitor()
            # Hook data
            raw = await self.async_load(compilers.endpoint(PORT_STATUS, self._identity))
            # Reset time
            monitor.reset()

            # Process data

            # Ports info
            ports = {
                LAN: dict(),
                USB: dict(),
                WAN: dict(),
            }
            data = raw[f"{PORT}_{INFO}"][self._identity.mac]
            for port in data:
                port_type = PORT_TYPES.get(port[0])
                port_id = converters.int_from_str(port[1:])
                # Replace needed key/value pairs
                for key in CONVERTERS[PORT_STATUS]:
                    if key.value in data[port]:
                        data[port][key.get()] = key.method(data[port][key.value])
                        data[port].pop(key.value)
                ports[port_type][port_id] = data[port]

            monitor[PORTS] = ports

            # Finish and save data
            monitor.finish()
            self.monitor[PORT_STATUS] = monitor

        except AsusRouterError as ex:
            self.monitor[PORT_STATUS].drop()
            raise ex

        return

    async def async_monitor_devices(self) -> None:
        """Monitor connected devices"""

        while self._monitor_devices.active:
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            return

        try:
            self._monitor_devices.start()
            monitor_devices = Monitor()

            # Update device list (needed for older devices)
            await self.async_check_endpoint(AR_PATH["devices_update"])
            if self._identity.update_networkmapd:
                await self.async_check_endpoint(AR_PATH["networkmap"])
            data = await self.async_hook(AR_HOOK_DEVICES)

            monitor_devices.reset()

            onboarding = {}
            if self._identity.onboarding:
                await self.async_monitor_onboarding()
                onboarding = self.monitor[ONBOARDING][CLIENTS]

            # Search for new data
            if AR_KEY_DEVICES in data:
                data = data[AR_KEY_DEVICES]
                for mac in data:
                    if converters.is_mac_address(mac):
                        monitor_devices[mac] = compilers.connected_device(
                            parsers.connected_device(data[mac]), onboarding.get(mac)
                        )
            # Or keep last data
            elif self._monitor_devices.ready:
                monitor_devices = self._monitor_devices

            monitor_devices.finish()
            self._monitor_devices = monitor_devices
        except AsusRouterError as ex:
            self._monitor_devices.drop()
            raise ex

        return

    ### <-- MONITORS

    ### TECHNICAL -->

    async def async_monitor_available(self, monitor: str) -> bool:
        """Check whether monitor is available"""

        # Monitor does not exist
        if not monitor in self.monitor:
            return False

        # Monitor is disabled
        if self.monitor[monitor].enabled != True:
            return False

        return True

    async def async_monitor_cached(self, monitor: str, value: Any) -> bool:
        """Check whether monitor has cached value"""

        now = datetime.utcnow()
        if (
            not self.monitor[monitor].ready
            or not value in self.monitor[monitor]
            or self._cache_time < (now - self.monitor[monitor].time).total_seconds()
        ):
            return False

        return True

    async def async_monitor_should_run(self, monitor: str) -> bool:
        """Check whether monitor should be run"""

        # Monitor is not available
        if await self.async_monitor_available(monitor) == False:
            return False

        # Monitor is already running - wait to complete
        if self.monitor[monitor].active:
            while self.monitor[monitor].active:
                await asyncio.sleep(DEFAULT_SLEEP_TIME)
            return False

        return True

    ### <-- TECHNICAL

    async def async_handle_error(self) -> None:
        """Actions to be taken on connection errors"""

        # Clear main monitor to prevent calculation errors in it
        if self._monitor_main.ready:
            self._monitor_main = Monitor()

        # Clear error
        await self.connection.async_reset_error()

        return

    async def async_handle_reboot(self) -> None:
        """Actions to be taken on reboot"""

        await self.async_keep_state_led()

        return

    async def async_find_interfaces(self, use_cache: bool = True) -> None:
        """Find available interfaces/type dictionary"""

        if not self._monitor_nvram.ready or use_cache == False:
            await self.async_monitor_nvram()

        ports = {}

        data = self._monitor_nvram

        for if_type in INTERFACE_TYPE:
            if if_type in data:
                values = data[if_type].split(" ")
                for item in values:
                    ports[item] = INTERFACE_TYPE[if_type]

        self._device_ports = ports

        return

    async def async_initialize(self):
        """Get all the data needed at the startup"""

        await self.async_monitor_main()
        await self.async_monitor_nvram()
        await self.async_find_interfaces()
        await self.async_monitor_misc()
        await self.async_monitor_devices()

    ### RETURN DATA -->

    async def async_get_aimesh(self, use_cache: bool = True) -> dict[str, AiMeshDevice]:
        """Return AiMesh device list"""

        monitor = ONBOARDING
        result = dict()

        # Monitor is not available
        if not await self.async_monitor_available(monitor):
            return result

        # Value is not cached or cache is disabled
        if not await self.async_monitor_cached(monitor, AIMESH) or use_cache == False:
            await self.async_monitor_onboarding()

        # Process data
        result = self.monitor[monitor][AIMESH]

        # Return data
        return result

    async def async_get_cpu(self, use_cache: bool = True) -> dict[str, float]:
        """Return CPU usage"""

        now = datetime.utcnow()
        if (
            not self._monitor_main.ready
            or not KEY_CPU in self._monitor_main
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
            or self._device_cpu_cores is None
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if CPU was monitored
        if not self._monitor_main.ready or not KEY_CPU in self._monitor_main:
            return result

        for core in self._device_cpu_cores:
            if DATA_USAGE in self._monitor_main[KEY_CPU][core]:
                result[DATA_BY_CORE.format(core)] = self._monitor_main[KEY_CPU][core][
                    DATA_USAGE
                ]
        if DATA_USAGE in self._monitor_main[KEY_CPU][DATA_TOTAL]:
            result[DATA_TOTAL] = self._monitor_main[KEY_CPU][DATA_TOTAL][DATA_USAGE]

        return result

    async def async_get_cpu_labels(self) -> list[str]:
        """Return list of CPU cores"""

        if self._device_cpu_cores is None:
            await self.async_monitor_main()

        result = list()
        result.append(DATA_TOTAL)
        for core in self._device_cpu_cores:
            result.append(DATA_BY_CORE.format(core))

        return result

    async def async_get_devicemap(self, use_cache: bool = True) -> dict[str, Any]:
        """Return devicemap"""

        now = datetime.utcnow()
        if (
            not self._monitor_devices.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time
                < (now - self._monitor_devices.time).total_seconds()
            )
        ):
            await self.async_monitor_misc()

        return self._monitor_misc["DEVICEMAP"]

    async def async_get_devices(self, use_cache: bool = True) -> dict[str, Any]:
        """Return device list"""

        now = datetime.utcnow()
        if (
            not self._monitor_devices.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time
                < (now - self._monitor_devices.time).total_seconds()
            )
        ):
            await self.async_monitor_devices()

        return self._monitor_devices.copy()

    async def async_get_gwlan(self) -> dict[str, Any]:
        """Get state of guest WLAN by id"""

        ids = await self.async_get_gwlan_ids()

        # NVRAM values to check
        nvram = list()
        for id in ids:
            for value in NVRAM_TEMPLATE["GWLAN"]:
                nvram.append(value.value.format(id))

        try:
            data = await self.async_hook(compilers.nvram(nvram))
        except AsusRouterError as ex:
            raise ex
        for id in ids:
            for value in NVRAM_TEMPLATE["GWLAN"]:
                data[value.value.format(id)] = value.method(
                    data[value.value.format(id)]
                )

        return data

    async def async_get_gwlan_ids(self) -> list[int]:
        """Return list of guest WLAN ids"""

        ids = list()

        wlans = await self.async_get_wlan_ids()

        for wlan in wlans:
            for i in range(1, 4):
                ids.append(f"{wlan}.{i}")

        return ids

    async def async_get_network(
        self, use_cache: bool = True
    ) -> dict[str, (int | float)]:
        """Return traffic and speed for all interfaces"""

        now = datetime.utcnow()
        if (
            not self._monitor_main.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if network was monitored
        if not self._monitor_main.ready or not KEY_NETWORK in self._monitor_main:
            return result

        for interface in self._monitor_main[KEY_NETWORK]:
            for value in self._monitor_main[KEY_NETWORK][interface]:
                result["{}_{}".format(interface, value)] = self._monitor_main[
                    KEY_NETWORK
                ][interface][value]

        return result

    async def async_get_network_labels(self) -> list[str]:
        """Return list of network interfaces"""

        if not self._monitor_main.ready:
            await self.async_monitor_main()
        if not self._identity:
            await self.async_identify()

        result = list()

        if not KEY_NETWORK in self._monitor_main:
            self._monitor_main.ready = False
            return await self.async_get_network_labels()
        for interface in self._monitor_main[KEY_NETWORK]:
            result.append(interface)

        if not "USB" in result and "dualwan" in self._identity.services:
            result.append("USB")

        return result

    ### PARENTAL CONTROL -->
    async def async_apply_parental_control_rules(
        self,
        rules: dict[str, FilterDevice],
    ) -> bool:
        """Apply parental control rules"""

        request = compilers.parental_control(rules)
        request["action_mode"] = "apply"

        return await self.async_service_run(
            service="restart_firewall",
            arguments=request,
        )

    async def async_get_parental_control(self) -> dict[str, Any]:
        """Return parental control status"""

        # NVRAM values to check
        nvram = list()
        nvram.append(AR_KEY_PARENTAL_CONTROL.value)
        for value in AR_MAP_PARENTAL_CONTROL:
            nvram.append(value)

        data = await self.async_hook(compilers.nvram(nvram))
        data = parsers.parental_control(data)

        return data

    async def async_get_parental_control_rules(
        self,
        macs: str | list[str] | None = None,
    ) -> dict[str, FilterDevice]:
        """Return parental control rules"""

        pc: dict = await self.async_get_parental_control()
        rules = pc.get("list", dict())

        # Return the full list if no device specified
        if macs is None:
            return rules

        # Return only devices of interest
        result = dict()
        macs = converters.as_list(macs)
        for mac in macs:
            if mac in rules:
                result[mac] = rules[mac]
        return result

    async def async_remove_parental_control_rules(
        self,
        macs: str | list[str] | None = None,
        rules: FilterDevice | list[FilterDevice] = None,
        apply: bool = True,
    ) -> dict[str, FilterDevice]:
        """Remove parental control rules"""

        macs = list() if macs is None else converters.as_list(macs)
        rules = list() if rules is None else converters.as_list(rules)

        # Get current rules
        cr: dict = await self.async_get_parental_control_rules()

        # Remove old rules for these MACs
        for mac in macs:
            if mac in cr:
                cr.pop(mac)
        for rule in rules:
            if rule.mac in cr:
                cr.pop(rule.mac)

        # Apply new rules
        if apply:
            await self.async_apply_parental_control_rules(cr)

        # Return the new rules
        return cr

    async def async_set_parental_control_rules(
        self,
        rules: FilterDevice | list[FilterDevice],
    ) -> bool:
        """Set parental control rules"""

        rules = converters.as_list(rules)

        # Remove old rules for these MACs and get the rest of the list
        cr = await self.async_remove_parental_control_rules(rules, apply=False)

        # Add new rules
        for rule in rules:
            cr[rule.mac] = rule

        # Apply new rules
        return await self.async_apply_parental_control_rules(cr)

    ### <-- PARENTAL CONTROL

    async def async_get_ports(
        self, use_cache: bool = True
    ) -> dict[str, dict[str, int]]:
        """Return WAN/LAN ports status"""

        now = datetime.utcnow()
        if (
            not self._monitor_misc.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_misc.time).total_seconds()
            )
        ):
            await self.async_monitor_misc()

        return self._monitor_misc["PORTS"]

    async def async_get_ram(self, use_cache: bool = True) -> dict[str, (int | float)]:
        """Return RAM and its usage"""

        now = datetime.utcnow()
        if (
            not self._monitor_main.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if RAM was monitored
        if not self._monitor_main.ready or not KEY_RAM in self._monitor_main:
            return result

        for value in self._monitor_main[KEY_RAM]:
            result[value] = self._monitor_main[KEY_RAM][value]

        return result

    async def async_get_sysinfo(self, use_cache: bool = True) -> dict[str, Any]:
        """Return sysinfo status"""

        if not self._identity.sysinfo:
            return {}

        now = datetime.utcnow()
        if (
            not self._monitor_main.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        if not KEY_SYSINFO in self._monitor_main:
            return {}

        return self._monitor_main[KEY_SYSINFO]

    async def async_get_temperature(self, use_cache: bool = True) -> dict[str, Any]:
        """Raturn temperature status"""

        now = datetime.utcnow()
        if (
            not self._monitor_misc.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_misc.time).total_seconds()
            )
        ):
            await self.async_monitor_misc()

        return self._monitor_misc[KEY_TEMPERATURE]

    async def async_get_vpn(self, use_cache: bool = True) -> dict[str, Any]:
        """Return VPN status"""

        now = datetime.utcnow()
        if (
            not self._monitor_misc.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_misc.time).total_seconds()
            )
        ):
            await self.async_monitor_misc()

        return self._monitor_misc[KEY_VPN]

    async def async_get_wan(self, use_cache: bool = True) -> dict[str, str]:
        """Return WAN and its usage"""

        now = datetime.utcnow()
        if (
            not self._monitor_main.ready
            or use_cache == False
            or (
                use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if WAN was monitored
        if not self._monitor_main.ready or not KEY_WAN in self._monitor_main:
            return result

        for value in self._monitor_main[KEY_WAN]:
            result[value] = self._monitor_main[KEY_WAN][value]

        return result

    async def async_get_wlan(self) -> dict[str, Any]:
        """Get state of WLAN by id"""

        ids = await self.async_get_wlan_ids()

        # NVRAM values to check
        nvram = list()
        for id in ids:
            for value in NVRAM_TEMPLATE["WLAN"]:
                nvram.append(value.value.format(id))

        try:
            data = await self.async_hook(compilers.nvram(nvram))
        except AsusRouterError as ex:
            raise ex
        for id in ids:
            for value in NVRAM_TEMPLATE["WLAN"]:
                data[value.value.format(id)] = value.method(
                    data[value.value.format(id)]
                )

        return data

    async def async_get_wlan_ids(self) -> list[int]:
        """Return list of WLAN ids"""

        ids = list()

        interfaces = await self.async_get_network_labels()
        for value in interfaces:
            if value[:4] == "WLAN":
                ids.append(int(value[-1:]))

        return ids

    ### <-- RETURN DATA

    ### SERVICE RUN -->

    async def async_service_run(
        self,
        service: str,
        arguments: dict[str, Any] | None = None,
        expect_modify: bool = True,
        drop_connection: bool = False,
    ) -> bool:
        """Generic service to run"""

        commands = {
            AR_KEY_SERVICE_COMMAND: service,
        }

        if arguments:
            commands.update(arguments)

        try:
            result = await self.async_command(commands=commands)
        except AsusRouterServerDisconnectedError as ex:
            # For services, that will block any further connections
            if drop_connection:
                _LOGGER.debug(MSG_INFO["drop_connection_service"].format(service))
                await self.async_drop_connection()
                return True
            else:
                raise ex

        if (
            not AR_KEY_SERVICE_REPLY in result
            or result[AR_KEY_SERVICE_REPLY] != service
            or not AR_KEY_SERVICE_MODIFY in result
        ):
            raise AsusRouterServiceError(ERROR_SERVICE.format(service, result))

        _LOGGER.debug(MSG_INFO["service"].format(service, arguments, result))

        services = service.split(";")
        if any(service in TRACK_SERVICES_LED for service in services):
            await self.async_keep_state_led()

        if expect_modify:
            return converters.bool_from_any(result[AR_KEY_SERVICE_MODIFY])
        return True

    ### SERVICES -->

    async def async_service_control(self, target: str, mode: str) -> bool:
        """Start / stop / (force) restart service"""

        if not target in AR_SERVICE_CONTROL or not mode in AR_SERVICE_CONTROL[target]:
            raise AsusRouterServiceError(ERROR_SERVICE_UNKNOWN.format(target, mode))

        service = AR_SERVICE_COMMAND[mode].format(target)

        drop_connection = False
        if target in AR_SERVICE_DROP_CONNECTION:
            drop_connection = True

        return await self.async_service_run(
            service=service, drop_connection=drop_connection
        )

    async def async_service_led_get(self) -> bool | None:
        """Return status of the LEDs"""

        key = AR_KEY_LED
        raw = await self.async_hook(compilers.nvram(key))
        if not converters.none_or_str(raw[key]):
            return None
        value = converters.bool_from_any(raw[key])
        self._status_led = value
        return value

    async def async_service_led_set(self, value: bool | int | str) -> bool:
        """Set status of the LEDs"""

        value_to_set = converters.bool_from_any(value)

        service = "start_ctrl_led"
        arguments = {
            AR_KEY_LED: converters.int_from_bool(value_to_set),
        }
        result = await self.async_service_run(service=service, arguments=arguments)
        if result:
            self._status_led = value_to_set
        return result

    async def async_service_ledg_get(self) -> dict[str, Any] | None:
        """Return status of RGB LEDs in LEDG scheme"""

        nvram = list()
        for mode in AR_LEDG_MODE:
            nvram.append(AR_KEY_LEDG_RGB.format(mode))
        nvram.append(AR_KEY_LEDG_SCHEME)
        nvram.append(AR_KEY_LEDG_SCHEME_OLD)

        data = await self.async_hook(compilers.nvram(nvram))

        ledg = dict()
        if AR_KEY_LEDG_SCHEME in data and data[AR_KEY_LEDG_SCHEME] != str():
            self._ledg_mode = data[AR_KEY_LEDG_SCHEME]
            ledg[AR_KEY_LEDG_SCHEME] = self._ledg_mode
        else:
            return None

        if AR_KEY_LEDG_RGB.format(self._ledg_mode) in data:
            self._ledg_color = parsers.rgb(
                data[AR_KEY_LEDG_RGB.format(self._ledg_mode)]
            )

        return {
            PARAM_COLOR: self._ledg_color,
            PARAM_COUNT: self._ledg_count,
            PARAM_MODE: self._ledg_mode,
        }

    async def async_service_ledg_set(
        self, mode: int, color: dict[int, dict[str, int]]
    ) -> bool:
        """Set state of RGB LEDs in LEDG scheme"""

        # If none LEDs
        if self._ledg_count == 0:
            return False

        if not mode in AR_LEDG_MODE:
            raise (AsusRouterValueError(ERROR_VALUE.format(mode)))

        # Check for the known state
        if not self._ledg_mode or not self._ledg_color:
            await self.async_service_ledg_get()

        colors = calculators.rgb(color)
        for num in range(1, self._ledg_count + 1):
            if not num in colors:
                if self._ledg_color and num in self._ledg_color:
                    colors[num] = self._ledg_color[num]
                else:
                    colors[num] = dict()

        color_to_set = compilers.rgb(colors)

        result = await self.async_load(
            f"{AR_PATH['ledg']}?{AR_KEY_LEDG_SCHEME}={mode}&ledg_rgb={color_to_set}"
        )

        if "statusCode" in result and int(result["statusCode"]) == 200:
            return True

        return False

    async def async_service_reboot(self) -> bool:
        """Reboot the device"""

        service = "reboot"
        await self.async_service_run(service=service, drop_connection=True)
        return True

    async def async_get_firmware_update(self) -> dict[str, Any]:
        """Check for firmware updates"""

        result = dict()

        # Check for updates
        await self.async_command(
            commands=None,
            action_mode=AR_FIRMWARE_CHECK_COMMAND,
            apply_to=AR_PATH["apply"],
        )
        # Get available firmware data
        data = await self.async_load(AR_PATH["firmware"])

        fw_current = self._identity.firmware
        fw_new = parsers.firmware_string(data["webs_state_info"])

        result["state"] = True if fw_current < fw_new else False
        result["current"] = str(fw_current)
        result["available"] = str(fw_new)

        return result

    ### <-- SERVICES

    async def async_keep_state_led(self) -> bool:
        """Keep state of LEDs"""

        # Only for Merlin firmware, so sysinfo should be present
        if self._identity.sysinfo and self.led == False:
            await self.async_service_led_set(True)
            await self.async_service_led_set(False)

    @property
    def connected(self) -> bool:
        return self.connection.connected

    @property
    def boottime(self) -> datetime:
        return self._device_boottime

    @property
    def led(self) -> bool:
        return self._status_led

    @property
    def ledg_count(self) -> int:
        return self._ledg_count



"""Convertors module for AsusRouter"""

from __future__ import annotations

import re
from datetime import datetime, timedelta, timezone
from dateutil.parser import parse as dtparse
from typing import Any

from asusrouter.error import AsusRouterValueError

BOOL_FALSE: tuple[str, ...] = (
    "false",
    "block",
    "0",
    "off",
    "disabled",
)
BOOL_TRUE: tuple[str, ...] = (
    "true",
    "allow",
    "1",
    "on",
    "enabled",
)
ERROR_VALUE = "Wrong value: {} with original exception: {}"
ERROR_VALUE_TYPE = "Wrong value {} of type {}"
REGEX_MAC = re.compile("^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})")

CONNECTION_TYPE = {
    "2G": 1,
    "5G": 2,
    "5G1": 3,
    "5G2": 3,
    "6G": 4,
}


def int_from_str(raw: str, base: int = 10) -> int:
    """Convert string to integer"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return int(0)

    try:
        return int(raw, base=base)
    except ValueError as ex:
        raise AsusRouterValueError(ERROR_VALUE.format(raw, str(ex)))


def clean_html(raw: str) -> str:
    """Clean html tags"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    return re.sub("<([\/ibu]+)>", "", raw)


def bool_or_int(raw: str, base: int = 10) -> bool | int:
    """Convert to bool or to int"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = clean_html(raw)

    try:
        return int_from_str(raw, base)
    except AsusRouterValueError:
        """Do nothing"""
    try:
        return bool_from_any(raw)
    except AsusRouterValueError:
        """Do nothing"""

    raise AsusRouterValueError(ERROR_VALUE.format(raw))


def float_from_str(raw: str) -> float:
    """Convert striing to float"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return float(0)
    # For devices reporrting speeds with M at the end for mega
    if raw[-1] == "M":
        raw = raw[:-1]

    try:
        value = float(raw)
    except ValueError as ex:
        raise AsusRouterValueError(ERROR_VALUE.format(raw, str(ex)))

    return value


def bool_from_any(raw: str | int | float | bool) -> bool | None:
    """Converts string or number to bool"""

    _type = type(raw)

    if _type == bool:
        return raw

    if _type == int or _type == float:
        if raw == 0:
            return False
        else:
            return True

    elif _type == str:
        if raw.lower().strip() in BOOL_FALSE:
            return False
        elif raw.lower().strip() in BOOL_TRUE:
            return True
        elif raw == str():
            return None
        else:
            raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    else:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))


def int_from_bool(raw: bool) -> int:
    """Convert boolean value to 0 or 1"""

    if type(raw) != bool:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    return 1 if raw else 0


def none_or_str(raw: str) -> str | None:
    """Returns either string or None if string is empty"""

    if raw is None:
        return None

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    if raw.strip() == str():
        return None

    return raw.strip()


def exists_or_not(raw: str) -> bool:
    """Returns whether value exists or is empty"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    if raw.strip() == str():
        return False

    return True


def timedelta_long(raw: str) -> timedelta:
    """Transform connection timedelta of the device to a proper datetime object when the device was connected"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return timedelta()

    part = raw.split(":")
    try:
        return timedelta(
            hours=int(part[-3]), minutes=int(part[-2]), seconds=int(part[-1])
        )
    except ValueError as ex:
        raise (AsusRouterValueError(ERROR_VALUE.format(raw, str(ex))))


def time_from_delta(raw: str) -> datetime:
    """Transform time delta to the date in the past"""

    return datetime.utcnow().replace(
        microsecond=0, tzinfo=timezone.utc
    ) - timedelta_long(raw)


def datetime_from_str(raw: str) -> datetime:
    """Transform string to datetime"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    raw = raw.strip()
    if raw == str():
        return {}

    try:
        return dtparse(raw)
    except Exception as ex:
        raise ex


def is_mac_address(raw: str) -> bool:
    """Checks if string is MAC address"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return False

    if re.search(REGEX_MAC, raw):
        return True

    return False


def service_support(raw: str) -> list[str]:
    """Get the list of the supported services"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    services = list()

    if raw == str():
        return services

    services = raw.split(" ")
    services = [i for i in services if i]

    return services


def ovpn_remote_fom_str(raw: str) -> dict[str, Any]:
    """Get OpenVPN remote data"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    try:
        temp = raw.split(",")
        auth = temp[1]
        temp = temp[0].split(":")
        ip = temp[0]
        port = int_from_str(temp[1])

        return {
            "ip": ip,
            "port": port,
            "auth": auth,
        }
    except Exception as ex:
        raise ex


def to_snake_case(raw: str):
    """Convert string to snake_case"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    string = (
        re.sub(r"(?<=[a-z])(?=[A-Z])|[^a-zA-Z]", " ", raw).strip().replace(" ", "_")
    )
    result = "".join(string.lower())
    while "__" in result:
        result = result.replace("__", "_")

    return result


def as_list(raw: Any):
    """Convert object into list of objects"""

    if type(raw) != list:
        return [raw]

    return raw


def onboarding_connection(raw: str) -> dict[str, int]:
    """Convert connection type from onboarding type to native type"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    if raw == "wired_mac":
        return {
            "connection_type": 0,
            "guest": 0,
        }

    try:
        temp = raw.split("_")
        return {
            "connection_type": CONNECTION_TYPE.get(temp[0]),
            "guest": temp[1] if len(temp) > 1 else 0,
        }
    except Exception as ex:
        raise ex



"""Dataclass module for AsusRouter"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

from asusrouter.util.converters import none_or_str

DEFAULT_PARENTAL_CONTROL_TIMEMAP = "W03E21000700<W04122000800"


@dataclass
class ConnectedDevice:
    """Connected device class"""

    name: str | None = None
    mac: str | None = None

    ip: str | None = None
    ip_method: str | None = None

    internet_state: bool | None = None
    internet_mode: bool | None = None

    connection_type: str | None = None
    node: str | None = None

    # WLAN only values
    online: bool | None = None
    rssi: int | None = None
    connected_since: int | None = None
    rx_speed: float | None = None
    tx_speed: float | None = None

    guest: int = 0


@dataclass
class AsusDevice:
    """Asus device class"""

    serial: str | None = None
    mac: str | None = None
    model: str | None = None
    brand: str = "ASUSTek"
    firmware: Firmware | None = None
    services: str | None = None
    sysinfo: bool = False
    led: bool = False
    ledg: bool = False
    aura: bool = False
    vpn_status: bool = False
    onboarding: bool = False
    port_status: bool = False
    update_networkmapd: bool = False
    ethernet_ports: bool = False


@dataclass
class AiMeshDevice:
    """AiMesh device class"""

    # Status
    status: bool = False

    alias: str | None = None
    model: str | None = None
    product_id: str | None = None
    ip: str | None = None

    fw: str | None = None
    fw_new: str | None = None

    mac: str | None = None

    # Access point: ap2g, ap5g, ap5g1, ap6g, apdwb
    ap: dict[str, Any] | None = None
    # Parent AiMesh: pap2g, rssi2g, pap2g_ssid, pap5g, rssi5g, pap5g_ssid, pap6g, rssi6g, pap6g_ssid
    parent: dict[str, Any] | None = None
    # Node state
    state: int | None = None
    level: int | None = None
    config: dict[str, Any] | None = None


@dataclass
class Key:
    """Key class"""

    value: str | int
    value_to_use: str = ""
    method: function = none_or_str

    def __str__(self) -> str:
        """Return only `value` as default"""

        return str(self.value)

    def get(self) -> str:
        """
        Get the proper value

        Returns
        -----
        `value_to_use` if exists, `value` otherwise
        """

        if self.value_to_use != "":
            return self.value_to_use

        return self.value


@dataclass
class Monitor(dict):
    """
    Monitor class

    In general this is dict with additions

    Properties
    -----
    `active`: bool flag of monitor being active

    `time`: datetime object showing the last time monitor was updated

    `ready`: bool flag if monitor was ever loaded

    Methods
    -----
    `start`: set `active` to True

    `stop`: set `active` to False

    `reset`: set `time` to utcnow()

    `finish`: set `ready` to True
    """

    active: bool = False
    time: datetime | None = None
    ready: bool = False
    enabled: bool = True

    def start(self) -> None:
        """Set to active"""

        self.active = True

    def stop(self) -> None:
        """Set to not-active"""

        self.active = False

    def reset(self) -> None:
        """Reset time to utcnow"""

        self.time = datetime.utcnow()

    def finish(self) -> None:
        """Set ready status to True"""

        self.ready = True

    def drop(self) -> None:
        """Drop on errors"""

        self.stop()
        self.ready = False


@dataclass
class Firmware:
    """Firmware class"""

    major: str = "3.0.0.4"
    minor: int | None = None
    build: int | None = None
    build_more: int | str | None = None

    def __lt__(self, other: Firmware) -> bool:
        """Define less-than"""

        if self.minor and other.minor and self.minor < other.minor:
            return True
        if self.build and other.build and self.build < other.build:
            return True
        if (
            type(self.build_more) == int
            and type(self.build_more) == type(other.build_more)
            and self.build_more < other.build_more
        ):
            return True
        return False

    def __str__(self) -> str:
        """Define string"""

        return f"{self.major}.{self.minor}.{self.build}_{self.build_more}"


@dataclass(frozen=True)
class FilterDevice:
    """Device filter class"""

    mac: str | None = None
    name: str | None = None
    state: int | None = None
    timemap: str = DEFAULT_PARENTAL_CONTROL_TIMEMAP



"""AsusRouter constants module"""

from asusrouter.dataclass import Key
from asusrouter.util.converters import (
    bool_from_any,
    bool_or_int,
    datetime_from_str,
    exists_or_not,
    float_from_str,
    int_from_str,
    ovpn_remote_fom_str,
    service_support,
    time_from_delta,
)

### API VALUES
AR_USER_AGENT = "asusrouter--DUTUtil-"
AR_API = [
    "Model_Name",
    "AiHOMEAPILevel",
    "Httpd_AiHome_Ver",
]

### VALUES
AIMESH = "aimesh"
CLIENTS = "clients"
CONNECTION_TYPE = "connection_type"
ENDPOINTS = "endpoints"
GUEST = "guest"
INFO = "info"
IP = "ip"
LAN = "lan"
LINK_RATE = "link_rate"
MAC = "mac"
NODE = "node"
ONBOARDING = "onboarding"
PORT = "port"
PORTS = "ports"
RSSI = "rssi"
STATE = "state"
SYSINFO = "sysinfo"
USB = "usb"
WAN = "wan"

### ASUS DATA TYPES
ETHERNET_PORTS = "ethernet_ports"
PORT_STATUS = "port_status"

### ENDPOINTS
ENDPOINT = {
    ETHERNET_PORTS: "ajax_ethernet_ports.asp",
    ONBOARDING: "ajax_onboarding.asp",
    PORT_STATUS: "get_port_status.cgi",
    SYSINFO: "ajax_sysinfo.asp",
}

ENDPOINT_ARGS = {
    PORT_STATUS: {
        MAC: f"{NODE}_{MAC}",
    },
}

### TYPES

PORT_TYPES = {
    "L": LAN,
    "U": USB,
    "W": WAN,
}

SPEED_TYPES = {
    "X": 0,
    "M": 100,
    "G": 1000,
    "Q": 2500,
}

### CONVERTERS

CONVERTERS = {
    PORT_STATUS: [
        Key("is_on", STATE, method=bool_from_any),
    ]
}

### CONSTANTS, DATA TYPES AND COMMON PARAMETERS
CONST_BITSINBYTE = 8
CONST_PERCENTS = 100
CONST_ZERO = 0.0

DATA_FREE = "free"
DATA_TOTAL = "total"
DATA_USAGE = "usage"
DATA_USED = "used"
DATA_RX = "rx"
DATA_TX = "tx"
DATA_TRAFFIC = [DATA_RX, DATA_TX]
DATA_BY_CORE = "core_{}"
DATA_ADD_SPEED = "{}_speed"

DEFAULT_CACHE_TIME = 5
DEFAULT_PORT = {
    "http": 80,
    "https": 8443,
}
DEFAULT_SLEEP_RECONNECT = 5
DEFAULT_SLEEP_CONNECTING = 1
DEFAULT_SLEEP_TIME = 0.1
DEFAULT_TRAFFIC_OVERFLOW = 4294967296
DEFAULT_USAGE_DIGITS = 2

PARAM_COLOR = "color"
PARAM_COUNT = "count"
PARAM_ERRNO = "errno"
PARAM_IP = "ip"
PARAM_MODE = "mode"
PARAM_RIP = "rip"
PARAM_STATE = "state"
PARAM_STATUS = "status"
PARAM_UNKNOWN = "unknown"

### OUR KEYS
KEY_CPU = "CPU"
KEY_NETWORK = "NETWORK"
KEY_RAM = "RAM"
KEY_SYSINFO = "SYSINFO"
KEY_TEMPERATURE = "TEMPERATURE"
KEY_VPN = "VPN"
KEY_WAN = "WAN"

# OUR REGEX
REGEX_VARIABLES = '([a-zA-Z0-9\_-]+)\s*=\s*"(.*?)"(?=;)'

# VALUES TO IGNORE
VALUES_TO_IGNORE = [str(), "None", "0.0.0.0"]

# Services to recover LED state
TRACK_SERVICES_LED = [
    "restart_wireless",
]


### ASUSWRT KEYS, MAPS AND VALUES
AR_DEFAULT_CORES = [1]
AR_DEFAULT_CORES_RANGE = range(1, 8)
AR_DEFAULT_LEDG = 8
AR_DEFAULT_OVPN_CLIENTS = 5
AR_DEFAULT_OVPN_SERVERS = 2

AR_DEVICE_ATTRIBUTES_LIST: tuple[Key, ...] = (
    Key("mac", "name"),
    Key("name", "name"),
    Key("nickName", "name"),
    Key("mac"),
    Key("ip"),
    Key("ipMethod", "ip_method"),
    Key("internetState", "internet_state", method=bool_from_any),
    Key("internetMode", "internet_mode"),
    Key("isWL", "connection_type", method=int_from_str),
    Key("isOnline", "online", method=bool_from_any),
    Key("rssi", method=int_from_str),
    Key("wlConnectTime", "connected_since", time_from_delta),
    Key("curRx", "rx_speed", method=float_from_str),
    Key("curTx", "tx_speed", method=float_from_str),
    Key("isGN", "guest", method=int_from_str),
)
AR_DEVICE_IDENTITY: tuple[Key, ...] = (
    Key("serial_no", "serial"),
    Key("label_mac", "mac"),
    Key("lan_hwaddr", "lan_mac"),
    Key("wan_hwaddr", "wan_mac"),
    Key("productid", "model"),
    Key("firmver", "fw_major"),
    Key("buildno", "fw_minor"),
    Key("extendno", "fw_build"),
    Key("rc_support", "services", method=service_support),
    Key("ss_support", "services", method=service_support),
    Key("led_val", "led", method=exists_or_not),
)

AR_ERROR = {
    "authorisation": 2,
    "credentials": 3,
    "try_again": 7,
    "logout": 8,
}

AR_KEY_AURARGB = "aurargb"
AR_KEY_CPU = "cpu_usage"
AR_KEY_CPU_ITEM = "cpu{}_{}"
AR_KEY_CPU_LIST: tuple[Key, ...] = (Key(DATA_TOTAL), Key(DATA_USAGE, DATA_USED))
AR_KEY_DEVICE_NAME_LIST = ["nickName", "name", "mac"]
AR_KEY_DEVICEMAP = "devicemap"
AR_KEY_DEVICES = "get_clientlist"
AR_KEY_DEVICES_LIST = "maclist"
AR_KEY_HEADER = "HEADER"
AR_KEY_LED = "led_val"
AR_KEY_LEDG_COUNT = "ledg_count"
AR_KEY_LEDG_SCHEME = "ledg_scheme"
AR_KEY_LEDG_SCHEME_OLD = "ledg_scheme_old"
AR_KEY_LEDG_RGB = "ledg_rgb{}"
AR_KEY_NETWORK = "netdev"
AR_KEY_NETWORK_GROUPS = {
    "INTERNET": "WAN",  # main WAN
    "INTERNET1": "USB",  # secondary WAN (USB modem / phone)
    "WIRED": "WIRED",  # wired connections
    "BRIDGE": "BRIDGE",  # bridge
    "WIRELESS0": "WLAN0",  # 2.4 GHz WiFi
    "WIRELESS1": "WLAN1",  # 5 GHz WiFi
    "WIRELESS2": "WLAN2",  # 5 GHz WiFi #2 (<-- check)
}
AR_KEY_NETWORK_ITEM = "{}_{}"
AR_KEY_OVPN = "{}{}_{}"  # client/server, id, type
AR_KEY_OVPN_STATUS = (
    Key("REMOTE", "remote", method=ovpn_remote_fom_str),
    Key("Updated", "datetime", method=datetime_from_str),
    Key("TUN/TAP read bytes", "tun_tap_read", method=int_from_str),
    Key("TUN/TAP write bytes", "tun_tap_write", method=int_from_str),
    Key("TCP/UDP read bytes", "tcp_udp_read", method=int_from_str),
    Key("TCP/UDP write bytes", "tcp_udp_write", method=int_from_str),
    Key("Auth read bytes", "auth_read", method=int_from_str),
    Key("pre-compress bytes", "pre_compress", method=int_from_str),
    Key("post-compress bytes", "post_compress", method=int_from_str),
    Key("pre-decompress bytes", "pre_decompress", method=int_from_str),
    Key("post-decompress bytes", "post_decompress", method=int_from_str),
)
AR_KEY_OVPN_STATUS_SERVER = (
    Key("CLIENT_LIST", "client_list"),
    Key("ROUTING_TABLE", "routing_table"),
)
AR_KEY_RAM = "memory_usage"
AR_KEY_RAM_ITEM = "mem_{}"
AR_KEY_RAM_LIST = [DATA_FREE, DATA_TOTAL, DATA_USED]
AR_KEY_SERVICE_COMMAND = "rc_service"
AR_KEY_SERVICE_MODIFY = "modify"
AR_KEY_SERVICE_REPLY = "run_service"
AR_KEY_VPN_CLIENT = "vpn_client"
AR_KEY_VPN_SERVER = "vpn_server"
AR_KEY_WAN = "wanlink_state"
AR_KEY_WAN_STATE = (
    Key("wanstate", "state"),
    Key("wansbstate", "bstate"),
    Key("wanauxstate", "aux"),
    Key("autodet_state"),
    Key("autodet_auxstate"),
    Key("wanlink_status", "status", method=bool_from_any),
    Key("wanlink_type", "ip_type"),
    Key("wanlink_ipaddr", "ip"),
    Key("wanlink_netmask", "mask"),
    Key("wanlink_gateway", "gateway"),
    Key("wanlink_dns", "dns"),
    Key("wanlink_lease", "lease"),
    Key("wanlink_expires", "expires"),
    Key("is_private_subnet", "private_subnet", method=int_from_str),
    Key("wanlink_xtype", "xtype"),
    Key("wanlink_xipaddr", "xip"),
    Key("wanlink_xnetmask", "xmask"),
    Key("wanlink_xgateway", "xgateway"),
    Key("wanlink_xdns", "xdns"),
    Key("wanlink_xlease", "xlease"),
    Key("wanlink_xexpires", "xexpires"),
)

AR_LEDG_MODE: dict[int, str] = {
    1: "Gradient",
    2: "Static",
    3: "Breathing",
    4: "Evolution",
    5: "Rainbow",
    6: "Wave",
    7: "Marquee",
}

AR_MAP_RGB: dict[int, str] = {
    0: "r",
    1: "g",
    2: "b",
}
AR_MAP_SYSINFO: dict[str, list[Key]] = {
    "conn_stats_arr": [
        Key(0, "conn_total", method=int_from_str),
        Key(1, "conn_active", method=int_from_str),
    ],
    "cpu_stats_arr": [
        Key(0, "load_avg_1", method=float_from_str),
        Key(1, "load_avg_5", method=float_from_str),
        Key(2, "load_avg_15", method=float_from_str),
    ],
    "mem_stats_arr": [
        Key(0, "ram_total", method=float_from_str),
        Key(1, "ram_free", method=float_from_str),
        Key(2, "buffers", method=float_from_str),
        Key(3, "cache", method=float_from_str),
        Key(4, "swap_size", method=float_from_str),
        Key(5, "swap_total", method=float_from_str),
        Key(6, "nvram_used", method=int_from_str),
        Key(7, "jffs"),
    ],
    "wlc_24_arr": [
        Key(0, "2ghz_clients_associated", method=bool_or_int),
        Key(1, "2ghz_clients_authorized", method=bool_or_int),
        Key(2, "2ghz_clients_authenticated", method=bool_or_int),
    ],
    "wlc_51_arr": [
        Key(0, "5ghz_clients_associated", method=bool_or_int),
        Key(1, "5ghz_clients_authorized", method=bool_or_int),
        Key(2, "5ghz_clients_authenticated", method=bool_or_int),
    ],
    "wlc_52_arr": [
        Key(0, "5ghz2_clients_associated", method=bool_or_int),
        Key(1, "5ghz2_clients_authorized", method=bool_or_int),
        Key(2, "5ghz2_clients_authenticated", method=bool_or_int),
    ],
}
AR_MAP_TEMPERATURE: dict[str, list[str]] = {
    "2ghz": [
        'curr_coreTmp_2_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_0_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl0_raw="([0-9.]+)&deg;C',
    ],
    "5ghz": [
        'curr_coreTmp_5_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_1_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl1_raw="([0-9.]+)&deg;C',
    ],
    "5ghz2": [
        'curr_coreTmp_52_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_2_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl2_raw="([0-9.]+)&deg;C',
    ],
    "6ghz": [
        'curr_coreTmp_3_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl3_raw="([0-9.]+)&deg;C',
    ],
    "cpu": ['curr_cpuTemp="([0-9.]+)"', 'curr_coreTmp_cpu="([0-9.]+)"'],
}
AR_KEY_PARENTAL_CONTROL_STATE = "MULTIFILTER_ENABLE"
AR_KEY_PARENTAL_CONTROL_NAME = "MULTIFILTER_DEVICENAME"
AR_KEY_PARENTAL_CONTROL_MAC = "MULTIFILTER_MAC"
AR_KEY_PARENTAL_CONTROL_TIMEMAP = "MULTIFILTER_MACFILTER_DAYTIME_V2"
AR_MAP_PARENTAL_CONTROL: dict[str, str] = {
    "MULTIFILTER_ENABLE": "enable",
    "MULTIFILTER_DEVICENAME": "name",
    "MULTIFILTER_MAC": "mac",
    "MULTIFILTER_MACFILTER_DAYTIME_V2": "time",
}
AR_MAP_PARENTAL_CONTROL_STATE = {
    "0": "disable",
    "1": "time",
    "2": "block",
}
AR_KEY_PARENTAL_CONTROL = Key(
    "MULTIFILTER_ALL", "parental_control", method=bool_from_any
)

AR_PATH = {
    "apply": "apply.cgi",
    "command": "applyapp.cgi",
    "devicemap": "ajax_status.xml",
    "devices_update": "update_clients.asp",
    "firmware": "detect_firmware.asp",
    "get": "appGet.cgi",
    "ledg": "set_ledg.cgi",
    "login": "login.cgi",
    "logout": "Logout.asp",
    "networkmap": "update_networkmapd.asp",
    "onboarding": "ajax_onboarding.asp",
    "ports": "ajax_ethernet_ports.asp",
    "rgb": "light_effect.html",
    "state": "state.js",
    "sysinfo": "ajax_sysinfo.asp",
    "temperature": "ajax_coretmp.asp",
    "vpn": "ajax_vpn_status.asp",
}

AR_VPN_STATUS = {
    -1: "error",
    0: "disconnected",
    1: "connecting",
    2: "connected",
}

### ASUSWRT SERVICES
AR_HOOK_TEMPLATE = "{}({});"
AR_HOOK_DEVICES = "{}()".format(AR_KEY_DEVICES)

AR_FIRMWARE_CHECK_COMMAND = "firmware_check"
AR_SERVICE_COMMAND: dict[str, str] = {
    "force_restart": "restart_{}_force",
    "start": "start_{}",
    "stop": "stop_{}",
    "restart": "restart_{}",
}
AR_SERVICE_CONTROL: dict[str, list[str]] = {
    "firewall": [
        "restart",
    ],
    "ftpd": [
        "force_restart",
        "start",
        "stop",
        "restart",
    ],
    "httpd": [
        "restart",
    ],
    "wireless": [
        "restart",
    ],
}
AR_SERVICE_DROP_CONNECTION: list[str] = {
    "httpd",
}

### ERRORS AND MESSAGES
ERROR_IDENTITY = "Cannot obtain identity from the device {}. Exception summary{}"
ERROR_PARSING = (
    "Failed parsing value '{}'. Please report this issue. Exception summary: {}"
)
ERROR_SERVICE = "Error calling service '{}'. Service did not return any expected value in the reply: {}"
ERROR_SERVICE_UNKNOWN = "Unknown service '{}' with mode '{}'"
ERROR_VALUE = "Wrong value '{}' with original exception: {}"
ERROR_VALUE_TYPE = "Wrong value '{}' of type {}"
ERROR_ZERO_DIVISION = "Zero division allert: {}"

MSG_ERROR = {
    "authorisation": "Session is not authorised",
    "cert_missing": "CA certificate is missing",
    "command": "Error sending a command to {}. Exception: {}",
    "credentials": "Wrong credentials",
    "disabled_control": "Device is connected in no-control mode. Sending commands is blocked",
    "disabled_monitor": "Device is connected in no-monitor mode. Sending hooks is blocked",
    "token": "Cannot recieve a token from device",
    "try_again": "Too many attempts",
    "unknown": "Unknown error code {}",
}
MSG_INFO = {
    "drop_connection_service": "Connection will be dropped on behalf of '{}' service",
    "error_flag": "Error flag found",
    "empty_reqquest": "This request is empty",
    "identifying": "Identifying the device",
    "json_fix": "Trying to fix JSON response",
    "monitor_sleep": "Monitor {} is already active -> sleep",
    "monitor_wakeup": "Monitor {} woke up -> closing",
    "no_cert": "No certificate provided. Using trusted",
    "no_cert_check": "Certificate won't be checked",
    "reconnect": "Reconnecting",
    "service": "Service '{}' was called with arguments '{}' successfully. Reply: {}",
    "xml": "Data is in XML",
}
MSG_SUCCESS = {
    "cert_found": "CA certificate found",
    "command": "Command was sent successfully",
    "hook": "Hook was sent successfully",
    "identity": "Identity collected",
    "load": "Page {} was loaded successfully",
    "login": "Login successful",
    "logout": "Logout successful",
}
MSG_WARNING = {
    "disconnected": "Disconnected by the device while quering '{}' with payload '{}'. Everything should recover by itself. If this warning appears regularly, you might need to decrease the number of simultaneous connections to your device.",
    "not_connected": "Not connected",
    "refused": "Connection refused",
}


INTERFACE_TYPE = {
    "wan_ifnames": "wan",
    "wl_ifnames": "wlan",
    "wl0_vifnames": "gwlan2",
    "wl1_vifnames": "gwlan5",
    "lan_ifnames": "lan",
}

NVRAM_TEMPLATE = {
    "WLAN": [
        Key("wl{}_auth_mode_x"),
        Key("wl{}_bw"),
        Key("wl{}_channel"),
        Key("wl{}_chanspec"),
        Key("wl{}_closed", method=bool_from_any),
        Key("wl{}_country_code"),
        Key("wl{}_crypto"),
        Key("wl{}_gmode_check"),
        Key("wl{}_maclist_x"),
        Key("wl{}_macmode"),
        Key("wl{}_mbo_enable"),
        Key("wl{}_mfp"),
        Key("wl{}_nmode_x"),
        Key("wl{}_optimizexbox_ckb"),
        Key("wl{}_radio", method=bool_from_any),
        Key("wl{}_radius_ipaddr"),
        Key("wl{}_radius_key"),
        Key("wl{}_radius_port"),
        Key("wl{}_ssid"),
        Key("wl{}_wpa_gtk_rekey"),
        Key("wl{}_wpa_psk"),
    ],
    "GWLAN": [
        Key("wl{}_akm"),
        Key("wl{}_ap_isolate"),
        Key("wl{}_auth"),
        Key("wl{}_auth_mode"),
        Key("wl{}_auth_mode_x"),
        Key("wl{}_bridge"),
        Key("wl{}_bss_enabled", method=bool_from_any),
        Key("wl{}_bss_maxassoc", method=int_from_str),
        Key("wl{}_bw_dl", method=int_from_str),  # Bandwidth limit download
        Key("wl{}_bw_enabled", method=bool_from_any),  # Bandwidth limit switch
        Key("wl{}_bw_ul", method=int_from_str),  # Bandwidth limit upload
        Key("wl{}_closed", method=bool_from_any),
        Key("wl{}_crypto"),
        Key("wl{}_expire", method=int_from_str),  # Expire time in s
        Key("wl{}_expire_tmp", method=int_from_str),  # Expire time left in s
        Key("wl{}_gn_wbl_enable"),
        Key("wl{}_gn_wbl_rule"),
        Key("wl{}_gn_wbl_type"),
        Key("wl{}_hwaddr"),  # MAC address
        Key("wl{}_ifname"),  # Interface name
        Key("wl{}_infra"),
        Key("wl{}_key"),
        Key("wl{}_key1"),
        Key("wl{}_key2"),
        Key("wl{}_key3"),
        Key("wl{}_key4"),
        Key("wl{}_lanaccess", method=bool_from_any),  # LAN access
        Key("wl{}_maclist"),
        Key("wl{}_macmode"),
        Key("wl{}_maxassoc", method=int_from_str),
        Key("wl{}_mbss"),
        Key("wl{}_mfp"),
        Key("wl{}_mode"),
        Key("wl{}_net_reauth", method=int_from_str),
        Key("wl{}_preauth"),
        Key("wl{}_radio", method=bool_from_any),
        Key("wl{}_radius_ipaddr"),
        Key("wl{}_radius_key"),
        Key("wl{}_radius_port", method=int_from_str),
        Key("wl{}_sae_anti_clog_threshold"),
        Key("wl{}_sae_groups"),
        Key("wl{}_sae_sync"),
        Key("wl{}_ssid"),  # SSID
        Key("wl{}_sta_retry_time"),
        Key("wl{}_sync_node", method=bool_from_any),  # Sync AiMesh nodes
        Key("wl{}_unit"),  # GWLAN unit id
        Key("wl{}_wep", method=bool_from_any),
        Key("wl{}_wep_x", method=bool_from_any),
        Key("wl{}_wfi_enable", method=bool_from_any),
        Key("wl{}_wfi_pinmode"),
        Key("wl{}_wme"),
        Key("wl{}_wme_bss_disable", method=bool_from_any),
        Key("wl{}_wpa_gtk_rekey"),
        Key("wl{}_wpa_psk"),  # Password
        Key("wl{}_wps_mode"),
    ],
}

NVRAM_LIST = {
    "DEVICE": [
        "productid",
        "serial_no",
        "label_mac",
        "model",
        "pci/1/1/ATE_Brand",
    ],
    "MODE": [
        "sw_mode",
        "wlc_psta",
        "wlc_express",
    ],
    "FIRMWARE": [
        "firmver",
        "buildno",
        "extendno",
    ],
    "FUNCTIONS": [
        "rc_support",
        "ss_support",
    ],
    "INTERFACES": [
        "wans_dualwan",
        "wan_ifnames",
        "wl_ifnames",
        "wl0_vifnames",
        "wl1_vifnames",
        "lan_ifnames",
        "br0_ifnames",
    ],
    "REBOOT": [
        "reboot_schedule",
        "reboot_schedule_enable",
        "reboot_time",
    ],
    "WAN0": [
        "link_wan",
        "wan0_auxstate_t",
        "wan0_dns",
        "wan0_enable",
        "wan0_expires",
        "wan0_gateway",
        "wan0_gw_ifname",
        "wan0_gw_mac",
        "wan0_hwaddr",
        "wan0_ipaddr",
        "wan0_is_usb_modem_ready",
        "wan0_primary",
        "wan0_proto",
        "wan0_realip_ip",
        "wan0_realip_state",
        "wan0_sbstate_t",
        "wan0_state_t",
        "wan0_upnp_enable",
    ],
    "WAN1": [
        "link_wan1",
        "wan1_auxstate_t",
        "wan1_dns",
        "wan1_enable",
        "wan1_gateway",
        "wan1_gw_ifname",
        "wan1_gw_mac",
        "wan1_hwaddr",
        "wan1_ipaddr",
        "wan1_is_usb_modem_ready",
        "wan1_primary",
        "wan1_proto",
        "wan1_realip_ip",
        "wan1_realip_state",
        "wan1_sbstate_t",
        "wan1_state_t",
        "wan1_upnp_enable",
    ],
    "LAN": [
        "lan_gateway",
        "lan_hwaddr",
        "lan_ifname",
        "lan_ifnames",
        "lan_ipaddr",
        "lan_netmask",
        "lan_proto",
        "lanports",
    ],
    "WLAN0": [
        "wl0_radio",
        "wl0_ssid",
        "wl0_chanspec",
        "wl0_closed",
        "wl0_nmode_x",
        "wl0_bw",
        "wl0_auth_mode_x",
        "wl0_crypto",
        "wl0_wpa_psk",
        "wl0_mfp",
        "wl0_mbo_enable",
        "wl0_country_code",
        "wl0_maclist_x",
        "wl0_macmode",
    ],
    "WLAN1": [
        "wl1_radio",
        "wl1_ssid",
        "wl1_chanspec",
        "wl1_closed",
        "wl1_nmode_x",
        "wl1_bw",
        "wl1_auth_mode_x",
        "wl1_crypto",
        "wl1_wpa_psk",
        "wl1_mfp",
        "wl1_mbo_enable",
        "wl1_country_code",
        "wl1_maclist_x",
        "wl1_macmode",
    ],
    "GWLAN01": [
        "wl0.1_bss_enabled",
        "wl0.1_ssid",
        "wl0.1_auth_mode_x",
        "wl0.1_crypto",
        "wl0.1_key",
        "wl0.1_wpa_psk",
        "wl0.1_lanaccess",
        "wl0.1_expire",
        "wl0.1_expire_tmp",
        "wl0.1_macmode",
        "wl0.1_mbss",
        "wl0.1_sync_mode",
    ],
    "GWLAN02": [
        "wl0.2_bss_enabled",
        "wl0.2_ssid",
        "wl0.2_auth_mode_x",
        "wl0.2_crypto",
        "wl0.2_key",
        "wl0.2_wpa_psk",
        "wl0.2_lanaccess",
        "wl0.2_expire",
        "wl0.2_expire_tmp",
        "wl0.2_macmode",
        "wl0.2_mbss",
        "wl0.2_sync_mode",
    ],
    "GWLAN03": [
        "wl0.3_bss_enabled",
        "wl0.3_ssid",
        "wl0.3_auth_mode_x",
        "wl0.3_crypto",
        "wl0.3_key",
        "wl0.3_wpa_psk",
        "wl0.3_lanaccess",
        "wl0.3_expire",
        "wl0.3_expire_tmp",
        "wl0.3_macmode",
        "wl0.3_mbss",
        "wl0.3_sync_mode",
    ],
    "GWLAN11": [
        "wl1.1_bss_enabled",
        "wl1.1_ssid",
        "wl1.1_auth_mode_x",
        "wl1.1_crypto",
        "wl1.1_key",
        "wl1.1_wpa_psk",
        "wl1.1_lanaccess",
        "wl1.1_expire",
        "wl1.1_expire_tmp",
        "wl1.1_macmode",
        "wl1.1_mbss",
        "wl1.1_sync_mode",
    ],
    "GWLAN12": [
        "wl1.2_bss_enabled",
        "wl1.2_ssid",
        "wl1.2_auth_mode_x",
        "wl1.2_crypto",
        "wl1.2_key",
        "wl1.2_wpa_psk",
        "wl1.2_lanaccess",
        "wl1.2_expire",
        "wl1.2_expire_tmp",
        "wl1.2_macmode",
        "wl1.2_mbss",
        "wl1.2_sync_mode",
    ],
    "GWLAN13": [
        "wl1.3_bss_enabled",
        "wl1.3_ssid",
        "wl1.3_auth_mode_x",
        "wl1.3_crypto",
        "wl1.3_key",
        "wl1.3_wpa_psk",
        "wl1.3_lanaccess",
        "wl1.3_expire",
        "wl1.3_expire_tmp",
        "wl1.3_macmode",
        "wl1.3_mbss",
        "wl1.3_sync_mode",
    ],
}

MONITOR_MAIN = {
    "cpu_usage": "appobj",
    "memory_usage": "appobj",
    "netdev": "appobj",
    "wanlink_state": "appobj",
}

PORT_TYPE = [
    "LAN",
    "WAN",
]

KEY_NVRAM_GET = "nvram_get"
KEY_ACTION_MODE = "action_mode"
KEY_HOOK = "hook"
DEFAULT_ACTION_MODE = "apply"


### DEVICEMAP

# These values are just stored directly in this order in the corresponding node
DEVICEMAP_BY_INDEX = {
    "WAN": {
        "wan": [
            "status",
            "sbstatus",
            "auxstatus",
        ],
    },
    "WAN0": {
        "first_wan": [
            "status",
            "sbstatus",
            "auxstatus",
        ],
    },
    "WAN1": {
        "second_wan": [
            "status",
            "sbstatus",
            "auxstatus",
        ],
    },
    "USB": {
        "usb": [
            "status",
        ],
    },
}


# Not implemented yet
# psta:wlc_state=0;wlc_state_auth=0;
DEVICEMAP_SPECIAL = {
    "WAN": {
        "psta": [
            "wlc_state",
            "wlc_state_auth",
        ],
    },
}

# These values are stored as "key=value"
DEVICEMAP_GENERAL = {
    "WAN": {
        "wan": [
            "monoClient",
            "wlc_state",
            "wlc_sbstate",
            "wifi_hw_switch",
            "ddnsRet",
            "ddnsUpdate",
            "wan_line_state",
            "wlan0_radio_flag",
            "wlan1_radio_flag",
            "wlan2_radio_flag",
            "data_rate_info_2g",
            "data_rate_info_5g",
            "data_rate_info_5g_2",
            "wan_diag_state",
            "active_wan_unit",
            "wlc0_state",
            "wlc1_state",
            "rssi_2g",
            "rssi_5g",
            "rssi_5g_2",
            "link_internet",
            "wlc2_state",
            "le_restart_httpd",
        ],
    },
    "VPN": {
        "vpn": [
            "vpnc_proto",
            "vpnc_state_t",
            "vpnc_sbstate_t",
            "vpn_client1_state",
            "vpn_client2_state",
            "vpn_client3_state",
            "vpn_client4_state",
            "vpn_client5_state",
            "vpnd_state",
            "vpn_client1_errno",
            "vpn_client2_errno",
            "vpn_client3_errno",
            "vpn_client4_errno",
            "vpn_client5_errno",
        ],
    },
    "SYS": {
        "sys": [
            "uptimeStr",
        ],
    },
    "QTN": {
        "qtn": [
            "qtn_state",
        ],
    },
    "USB": {
        "usb": [
            "modem_enable",
        ],
    },
    "WAN0": {
        "wan": [
            "wan0_enable",
            "wan0_realip_state",
            "wan0_ipaddr",
            "wan0_realip_ip",
        ],
    },
    "WAN1": {
        "wan": [
            "wan1_enable",
            "wan1_realip_state",
            "wan1_ipaddr",
            "wan1_realip_ip",
        ],
    },
    "SIM": {
        "sim": [
            "sim_state",
            "sim_signal",
            "sim_operation",
            "sim_isp",
            "roaming",
            "roaming_imsi",
            "sim_imsi",
            "g3err_pin",
            "pin_remaining_count",
            "modem_act_provider",
            "rx_bytes",
            "tx_bytes",
            "modem_sim_order",
        ],
    },
    "DHCP": {
        "dhcp": [
            "dnsqmode",
        ],
    },
    "DIAG": {
        "diag": [
            "diag_dblog_enable",
            "diag_dblog_remaining",
        ]
    },
}

# This should be cleared from the data
DEVICEMAP_CLEAR = {
    "WAN": {
        "data_rate_info_2g": '"',
        "data_rate_info_5g": '"',
        "data_rate_info_5g_2": '"',
    }
}



"""Library init"""

from .error import (
    AsusRouter404,
    AsusRouterAuthorizationError,
    AsusRouterConnectionError,
    AsusRouterConnectionTimeoutError,
    AsusRouterError,
    AsusRouterIdentityError,
    AsusRouterLoginBlockError,
    AsusRouterLoginError,
    AsusRouterNotImplementedError,
    AsusRouterResponseError,
    AsusRouterServerDisconnectedError,
    AsusRouterServiceError,
    AsusRouterSSLError,
    AsusRouterValueError,
)
from .dataclass import (
    AiMeshDevice,
    AsusDevice,
    ConnectedDevice,
    FilterDevice,
    Firmware,
    Key,
    Monitor,
)
from .connection import Connection
from .asusrouter import AsusRouter



"""Parsers module for AsusRouter"""

from __future__ import annotations

import logging
from time import sleep

_LOGGER = logging.getLogger(__name__)

import json
import re
from datetime import datetime, timedelta
from dateutil.parser import parse as dtparse
from typing import Any

import xmltodict

from asusrouter.const import (
    AR_DEFAULT_CORES,
    AR_DEFAULT_CORES_RANGE,
    AR_DEFAULT_LEDG,
    AR_DEFAULT_OVPN_CLIENTS,
    AR_DEFAULT_OVPN_SERVERS,
    AR_DEVICE_ATTRIBUTES_LIST,
    AR_KEY_CPU_ITEM,
    AR_KEY_CPU_LIST,
    AR_KEY_DEVICEMAP,
    AR_KEY_HEADER,
    AR_KEY_NETWORK_GROUPS,
    AR_KEY_NETWORK_ITEM,
    AR_KEY_OVPN,
    AR_KEY_OVPN_STATUS,
    AR_KEY_OVPN_STATUS_SERVER,
    AR_KEY_PARENTAL_CONTROL,
    AR_KEY_PARENTAL_CONTROL_MAC,
    AR_KEY_PARENTAL_CONTROL_NAME,
    AR_KEY_PARENTAL_CONTROL_STATE,
    AR_KEY_PARENTAL_CONTROL_TIMEMAP,
    AR_KEY_RAM_ITEM,
    AR_KEY_RAM_LIST,
    AR_KEY_VPN_CLIENT,
    AR_KEY_VPN_SERVER,
    AR_KEY_WAN_STATE,
    AR_MAP_PARENTAL_CONTROL_STATE,
    AR_MAP_RGB,
    AR_MAP_SYSINFO,
    AR_MAP_TEMPERATURE,
    AR_PATH,
    CONST_BITSINBYTE,
    CONST_ZERO,
    DATA_ADD_SPEED,
    DATA_TOTAL,
    DATA_TRAFFIC,
    DEVICEMAP_BY_INDEX,
    DEVICEMAP_CLEAR,
    DEVICEMAP_GENERAL,
    ERROR_PARSING,
    ERROR_VALUE,
    ERROR_VALUE_TYPE,
    KEY_NETWORK,
    PARAM_IP,
    PARAM_RIP,
    PARAM_STATE,
    PARAM_STATUS,
    REGEX_VARIABLES,
    VALUES_TO_IGNORE,
)
from asusrouter.dataclass import AiMeshDevice, ConnectedDevice, FilterDevice, Firmware
from asusrouter.error import AsusRouterNotImplementedError, AsusRouterValueError
from asusrouter.util import calculators, converters


def cpu_cores(raw: dict[str, Any] | None = None) -> list[int]:
    """CPU cores parser"""

    cores = list()

    if raw is None:
        return cores

    for i in AR_DEFAULT_CORES_RANGE:
        if any(
            AR_KEY_CPU_ITEM.format(i, data_type) in raw for data_type in AR_KEY_CPU_LIST
        ):
            cores.append(i)
        else:
            break

    return cores


def cpu_usage(
    raw: dict[str, Any], cores: list[int] = AR_DEFAULT_CORES
) -> dict[str, Any]:
    """CPU usage parser"""

    cpu = dict()

    # Populate total
    cpu[DATA_TOTAL] = dict()
    for item in AR_KEY_CPU_LIST:
        cpu[DATA_TOTAL][item.get()] = CONST_ZERO

    # Data / core
    for core in cores:
        cpu[core] = dict()

        for item in AR_KEY_CPU_LIST:
            key = AR_KEY_CPU_ITEM.format(core, item)
            new_key = item.get()
            if key in raw:
                try:
                    cpu[core][new_key] = int(raw[key])
                except ValueError as ex:
                    raise (AsusRouterValueError(ERROR_VALUE.format(raw[key], str(ex))))
                # Add this to total as well
                cpu[DATA_TOTAL][new_key] += cpu[core][new_key]

    return cpu


def ram_usage(raw: dict[str, Any]) -> dict[str, Any]:
    """RAM usage parser"""

    ram = dict()

    for item in AR_KEY_RAM_LIST:
        if AR_KEY_RAM_ITEM.format(item) in raw:
            try:
                ram[item] = int(raw[AR_KEY_RAM_ITEM.format(item)])
            except ValueError as ex:
                raise (
                    AsusRouterValueError(
                        ERROR_VALUE.format(raw[AR_KEY_RAM_ITEM.format(item)], str(ex))
                    )
                )

    return ram


def network_usage(
    raw: dict[str, Any], cache: dict[str, Any] | None = None
) -> dict[str, Any]:
    """Network usage parser"""

    network = dict()
    for group in AR_KEY_NETWORK_GROUPS:
        for type in DATA_TRAFFIC:
            if AR_KEY_NETWORK_ITEM.format(group, type) in raw:
                if not AR_KEY_NETWORK_GROUPS[group] in network:
                    network[AR_KEY_NETWORK_GROUPS[group]] = dict()
                try:
                    network[AR_KEY_NETWORK_GROUPS[group]][type] = int(
                        raw[AR_KEY_NETWORK_ITEM.format(group, type)], base=16
                    )
                except ValueError as ex:
                    raise (
                        AsusRouterValueError(
                            ERROR_VALUE.format(
                                raw[AR_KEY_NETWORK_ITEM.format(group, type)], str(ex)
                            )
                        )
                    )

            elif (
                cache is not None
                and KEY_NETWORK in cache
                and AR_KEY_NETWORK_GROUPS[group] in cache[DATA_TRAFFIC]
                and type in cache[DATA_TRAFFIC][AR_KEY_NETWORK_GROUPS[group]]
            ):
                network[AR_KEY_NETWORK_GROUPS[group]][type] = cache[DATA_TRAFFIC][
                    AR_KEY_NETWORK_GROUPS[group]
                ][type]

    return network


def network_speed(
    after: dict[str, dict[str, float]],
    before: dict[str, dict[str, float]],
    time_delta: float,
) -> dict[str, dict[str, float]]:
    """
    Network speed calculator

    Parameters
    -----
    `after`: current values. Outer dictionary `(groups)` contains inner dictionary `(types)`. On each `(type)` calculations are performed

    `before`: previous values

    `time_delta`: time between measurements

    Returns
    -----
    `after` with speeds append to `(types)`
    """

    for group in after:
        if group in before:
            speed = dict()
            for type in after[group]:
                if type in before[group]:
                    speed[
                        DATA_ADD_SPEED.format(type)
                    ] = CONST_BITSINBYTE * calculators.speed(
                        after=after[group][type],
                        before=before[group][type],
                        time_delta=time_delta,
                    )
                else:
                    speed[DATA_ADD_SPEED.format(type)] = CONST_ZERO
            after[group] |= speed

    return after


def wan_state(raw: dict[str, Any]) -> dict[str, Any]:
    """WAN status parser"""

    values = dict()

    for key in AR_KEY_WAN_STATE:
        if key.value in raw and raw[key.value] != str():
            try:
                values[key.get()] = (
                    key.method(raw[key.value]) if key.method else raw[key.value]
                )
            except AsusRouterValueError as ex:
                _LOGGER.warning(ERROR_PARSING.format(key.value, str(ex)))

    return values


def connected_device(raw: dict[str, Any]) -> ConnectedDevice:
    """Device parser"""

    values = dict()

    for key in AR_DEVICE_ATTRIBUTES_LIST:
        if key.value in raw and raw[key.value] != str():
            try:
                values[key.get()] = (
                    key.method(raw[key.value]) if key.method else raw[key.value]
                )
            except AsusRouterValueError as ex:
                _LOGGER.warning(ERROR_PARSING.format(key.value, str(ex)))

    device = ConnectedDevice(**values)

    return device


def uptime(data: str) -> datetime:
    """Uptime -> boot time parser"""

    try:
        part = data.split("(")
        seconds = int(re.search("([0-9]+)", part[1]).group())
        when = dtparse(part[0])
        boot = when - timedelta(seconds=seconds)
    except ValueError as ex:
        raise (AsusRouterValueError(ERROR_VALUE.format(data, str(ex))))

    return boot


def port_speed(value: str | None = None) -> int | None:
    """Port speed -> Mb/s parcer"""

    if value is None:
        return None
    elif value == "X":
        return 0
    elif value == "M":
        return 100
    elif value == "G":
        return 1000
    elif value == "Q":
        return 2500
    else:
        raise (AsusRouterNotImplementedError(value))


def devicemap(devicemap: dict[str, Any]) -> dict[str, Any]:
    """Devicemap parser"""

    data = {}

    # Get values only with index
    for node in DEVICEMAP_BY_INDEX:
        _node = {}
        for key in DEVICEMAP_BY_INDEX[node]:
            for el in range(len(DEVICEMAP_BY_INDEX[node][key])):
                _node[DEVICEMAP_BY_INDEX[node][key][el]] = devicemap[key][el]
            del devicemap[key][0 : (len(DEVICEMAP_BY_INDEX[node][key]) - 1)]
        data[node] = _node

    # Get values by key
    for node in DEVICEMAP_GENERAL:
        _node = {}
        for key in DEVICEMAP_GENERAL[node]:
            for el in DEVICEMAP_GENERAL[node][key]:
                if key in devicemap:
                    if type(devicemap[key]) == str:
                        if el in devicemap[key]:
                            _node[el] = devicemap[key].replace("{}=".format(el), "")
                            break
                    else:
                        for value in devicemap[key]:
                            if el in value:
                                _node[el] = value.replace("{}=".format(el), "")
                                break
        if node in data:
            data[node].update(_node)
        else:
            data[node] = _node

    # Clear values from useless symbols
    for node in DEVICEMAP_CLEAR:
        for value in DEVICEMAP_CLEAR[node]:
            data[node][value] = data[node][value].replace(
                DEVICEMAP_CLEAR[node][value], ""
            )

    return data


def ledg_count(raw: str) -> int:
    """LEDG count parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    value = re.search("ledg_count: ([0-9]+)", raw)
    if value:
        count = int(value[1])
    else:
        count = 0

    return count


def temperatures(raw: str) -> dict[str, Any]:
    """Temperature parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    temp = dict()

    for sensor in AR_MAP_TEMPERATURE:
        for reg in AR_MAP_TEMPERATURE[sensor]:
            value = re.search(reg, raw)
            if value:
                temp[sensor] = float(value[1])

    return temp


def rgb(raw: str, num: int = AR_DEFAULT_LEDG) -> dict[int, dict[str, int]]:
    """RGB to channels parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    leds = dict()

    color = re.findall("([0-9]+)", raw)
    length = len(color)

    for led in range(1, num):
        ind = (led - 1) * 3
        if ind + 2 > length:
            break
        if not led in leds:
            leds[led] = dict()
        for channel in AR_MAP_RGB:
            leds[led][AR_MAP_RGB[channel]] = color[ind + channel]

    return leds


def sysinfo(raw: str) -> dict[str, Any]:
    """Sysinfo parser"""

    raw = raw.replace("=", '":')
    raw = raw.replace(";", ',"')
    raw = '{"' + raw[:-2] + "}"
    data = json.loads(raw)

    result = dict()

    for set in AR_MAP_SYSINFO:
        if set in data:
            for key in AR_MAP_SYSINFO[set]:
                try:
                    result[key.get()] = (
                        key.method(data[set][key.value])
                        if key.method
                        else data[set][key.value]
                    )
                except AsusRouterValueError as ex:
                    _LOGGER.warning(ERROR_PARSING.format(key.value, str(ex)))

    return result


def onboarding(raw: str) -> dict[str, Any]:
    """Onboarding parser"""

    raw = raw.replace("\ufeff", "")
    raw = raw.replace("=", '":')
    raw = raw.replace(";", ',"')
    raw = raw.replace("[0]", "")
    raw = '{"' + raw[:-2] + "}"
    data = json.loads(raw)

    result = dict()

    return data


def aimesh_node(raw: dict[str, Any]) -> AiMeshDevice:
    """AiMesh node parser"""

    AP = {
        "2g": "2ghz",
        "5g": "5ghz",
        "5g1": "5ghz2",
        "6g": "6ghz",
        "dwb": "dwb",
    }

    FREQUENCIES = [
        "2g",
        "5g",
        "6g",
    ]

    ap = dict()
    for el in AP:
        if f"ap{el}" in raw and raw[f"ap{el}"] is not str():
            ap[AP[el]] = raw[f"ap{el}"]

    parent = dict()
    for el in FREQUENCIES:
        if f"pap{el}" in raw and raw[f"pap{el}"] is not str():
            parent["connection"] = AP[el]
            parent["mac"] = raw[f"pap{el}"]
            parent["rssi"] = converters.none_or_str(raw.get(f"rssi{el}"))
            parent["ssid"] = converters.none_or_str(raw.get(f"pap{el}_ssid"))

    level = converters.int_from_str(raw.get("level", "0"))
    state = "router" if level == 0 else "node"

    return AiMeshDevice(
        status=converters.bool_from_any(raw.get("online", 0)),
        alias=raw.get("alias", None),
        model=raw.get("ui_model_name", raw.get("model_name", None)),
        product_id=raw.get("product_id"),
        ip=raw.get("ip"),
        fw=raw.get("fwver", None),
        fw_new=converters.none_or_str(raw.get("newfwver")),
        mac=raw.get("mac", None),
        ap=ap,
        parent=parent,
        state=state,
        level=level,
        config=raw.get("config"),
    )


def parental_control(raw: dict[str, Any]) -> dict[str, Any]:
    """Parental control parser"""

    result = dict()

    result[AR_KEY_PARENTAL_CONTROL.get()] = AR_KEY_PARENTAL_CONTROL.method(
        raw[AR_KEY_PARENTAL_CONTROL.value]
    )

    # If no rules are set
    if raw.get(AR_KEY_PARENTAL_CONTROL_MAC) == raw.get(AR_KEY_PARENTAL_CONTROL_STATE):
        result["list"] = dict()
        return result

    list = dict()

    for el in raw:
        raw[el] = raw[el].split("&#62")

    size = len(raw["MULTIFILTER_MAC"])

    for i in range(0, size):
        device = FilterDevice(
            mac=raw[AR_KEY_PARENTAL_CONTROL_MAC][i],
            name=raw[AR_KEY_PARENTAL_CONTROL_NAME][i],
            state=AR_MAP_PARENTAL_CONTROL_STATE.get(
                raw[AR_KEY_PARENTAL_CONTROL_STATE][i], None
            ),
            timemap=raw[AR_KEY_PARENTAL_CONTROL_TIMEMAP][i]
            if i < len(raw[AR_KEY_PARENTAL_CONTROL_TIMEMAP])
            else "",
        )
        list[raw[AR_KEY_PARENTAL_CONTROL_MAC][i]] = device

    result["list"] = list.copy()

    return result


def pseudo_json(text: str, page: str) -> dict[str, Any]:
    """JSON parser"""

    if page == AR_PATH["vpn"]:
        return vpn_status(text)
    if page == AR_PATH["firmware"]:
        return firmware(text)

    data = re.sub("\s+", "", text)

    if "curr_coreTmp" in data:
        return temperatures(data)
    # Merlin v380
    elif page == AR_PATH["sysinfo"]:
        return sysinfo(data)
    elif page == AR_PATH["onboarding"]:
        return onboarding(data)
    elif "get_clientlist" in data:
        data = data.replace('"get_clientlist":', '"get_clientlist":{')
        data += "}"
    elif "get_wan_lan_status=" in data:
        data = data.replace("get_wan_lan_status=", "")
        data = data[:-1]
    elif "varcpuInfo,memInfo=newObject();cpuInfo=" in data:
        data = data.replace("varcpuInfo,memInfo=newObject();cpuInfo=", '{"cpu":{')
        data = data.replace(";memInfo=", '},"memory":{')
        data = data.replace(";", "}}")
    else:
        _LOGGER.error("Unknown data. Template for this data is unknown")
        return {}

    return json.loads(data.encode().decode("utf-8-sig"))


def xml(text: str, page: str) -> dict[str, Any]:
    """XML parser"""

    data = xmltodict.parse(text)
    if AR_KEY_DEVICEMAP in data:
        return devicemap(data[AR_KEY_DEVICEMAP])

    return {}


def ovpn_client_status(raw: str) -> dict[str, Any]:
    """OpenVPN client status parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    raw = raw.strip()
    if raw == str() or raw == "None":
        return {}

    values = dict()

    for key in AR_KEY_OVPN_STATUS:
        if key.value in raw:
            value = re.search(f"{key.value},(.*?)(?=>)", raw)
            if value:
                values[key.get()] = key.method(value[1]) if key.method else value[1]

    return values


def ovpn_server_status(raw: str) -> dict[str, Any]:
    """OpenVPN server status parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    raw = raw.strip()
    if raw == str() or raw == "None":
        return {}

    values = dict()

    for key in AR_KEY_OVPN_STATUS_SERVER:
        if key.value in raw:
            flag = False
            # Find headers
            header = re.search(f"{AR_KEY_HEADER},{key.value},(.*?)(?=>)", raw)
            if header:
                keys_raw = header[1].split(",")
                keys = list()
                for el in keys_raw:
                    keys.append(converters.to_snake_case(el))
                flag = True

            # Hide header
            raw = raw.replace(f"{AR_KEY_HEADER},{key.value},", "USED-")

            values[key.get()] = list()
            # Find values
            while value := re.search(f"{key.value},(.*?)(?=>)", raw):
                if not value:
                    break
                if flag:
                    cut = value[1].split(",")
                    i = 0
                    set = dict()
                    while i < len(cut) and i < len(keys):
                        set[keys[i]] = cut[i]
                        i += 1
                    values[key.get()].append(set.copy())
                else:
                    if key.method:
                        values[key.get()].append(key.method(value[1]))
                    else:
                        values[key.get()].append(value[1])
                raw = raw.replace(f"{key.value},{value[1]}", "USED-")

    return values


# var = "value";
def variables(raw: str) -> dict[str, Any]:
    """Variables parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    values = dict()
    temp = raw
    while len(temp) > 0:
        value = re.search(REGEX_VARIABLES, temp)
        if not value:
            break
        values[value[1]] = value[2]
        temp = temp.replace(f"{value[0]};", "")

    return values


def firmware(raw: str) -> dict[str, Any]:
    """Firmware parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    values = variables(raw.replace("'", '"'))

    return values


def firmware_string(raw: str) -> Firmware:
    """Firmware string parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    string = re.match(
        "^(3.?0.?0.?4)?[_.]?([0-9]{3})[_.]?([0-9]+)[_.-]?([a-zA-Z0-9-_]+)?$", raw
    )
    if not string:
        _LOGGER.warning(
            f"Firmware version cannot be parser. Please report this. The original FW string is: {raw}"
        )
        return Firmware()

    major = string[1]
    if major and not "." in major and len(major) == 4:
        major = major[0] + "." + major[1] + "." + major[2] + "." + major[3]

    minor = int(string[2])
    build = int(string[3])
    if string[4] and string[4].isdigit():
        build_more = int(string[4])
    else:
        build_more = string[4]

    fw = Firmware(
        minor=minor,
        build=build,
        build_more=build_more,
    )
    if major:
        fw.major = major

    return fw


def vpn_status(raw: str) -> dict[str, Any]:
    """VPN status parser"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    vpns = dict()

    values = variables(raw)

    # Clients
    for num in range(1, AR_DEFAULT_OVPN_CLIENTS + 1):
        key = AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_STATUS)
        if not key in values:
            break

        vpn_id = AR_KEY_VPN_CLIENT + str(num)

        if values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id] = ovpn_client_status(values[key])
            vpns[vpn_id][PARAM_STATE] = True
        else:
            vpns[vpn_id] = dict()
            vpns[vpn_id][PARAM_STATE] = False

        key = AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_IP)
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_IP] = values[key]

        key = AR_KEY_OVPN.format(AR_KEY_VPN_CLIENT, num, PARAM_RIP)
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_RIP] = values[key]

    # Servers
    for num in range(1, AR_DEFAULT_OVPN_SERVERS + 1):
        key = AR_KEY_OVPN.format(AR_KEY_VPN_SERVER, num, PARAM_STATUS)
        if not key in values:
            break

        vpn_id = AR_KEY_VPN_SERVER + str(num)

        if values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id] = ovpn_server_status(values[key])
            vpns[vpn_id][PARAM_STATE] = True
        else:
            vpns[vpn_id] = dict()
            vpns[vpn_id][PARAM_STATE] = False

        key = AR_KEY_OVPN.format(AR_KEY_VPN_SERVER, num, PARAM_IP)
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_IP] = values[key]

        key = AR_KEY_OVPN.format(AR_KEY_VPN_SERVER, num, PARAM_RIP)
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_RIP] = values[key]

    return vpns



"""Error module"""

from __future__ import annotations

import asyncio
from typing import Any, Optional

import aiohttp


class AsusRouterError(Exception):
    """Base class for errors in AsusRouter library"""

    def __init__(
        self, *args: Any, message: Optional[str] = None, **_kwargs: Any
    ) -> None:
        """Initialise base error"""

        super().__init__(*args, message)


class AsusRouterAuthorizationError(AsusRouterError):
    """Authorization error"""


class AsusRouterConnectionError(AsusRouterError, aiohttp.ClientConnectionError):
    """Error connecting to the router"""


class AsusRouterConnectionTimeoutError(
    AsusRouterError, aiohttp.ServerTimeoutError, asyncio.TimeoutError
):
    """Timeout error on communication"""


class AsusRouterServerDisconnectedError(AsusRouterError):
    """Server disconnected error"""


class AsusRouterSSLError(AsusRouterError):
    """SSL error"""


class AsusRouterLoginError(AsusRouterError):
    """Login error / credentials error"""


class AsusRouterLoginBlockError(AsusRouterError):
    """Too many attempts error on device side"""

    def __init__(
        self,
        *args: Any,
        message: Optional[str] = None,
        timeout: Optional[int] = None,
        **_kwargs: Any,
    ) -> None:
        """Initialise base error"""

        self._timeout = timeout

        super().__init__(*args, message)

    @property
    def timeout(self) -> int | None:
        """Return timeout"""

        return self._timeout


class AsusRouterResponseError(AsusRouterError):
    """Error on communication"""


class AsusRouterServiceError(AsusRouterError):
    """Error on calling a service"""


class AsusRouterValueError(AsusRouterError, ValueError):
    """Invalid value received"""


class AsusRouterNotImplementedError(AsusRouterError, NotImplementedError):
    """Not implemented error"""


class AsusRouterIdentityError(AsusRouterError):
    """Error of collecting device identity"""


class AsusRouter404(AsusRouterError):
    """Error on page not found"""
